#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-03-25
# @Filename: overwatcher.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import pathlib

from typing import Annotated, Literal

from fastapi import APIRouter, Path, Query
from pydantic import BaseModel, Field

from lvmapi.auth import AuthDependency
from lvmapi.tasks import get_gort_log_task
from lvmapi.tools.rabbitmq import CluClient


class OverwatcherStatusModel(BaseModel):
    """Overwatcher status model."""

    running: Annotated[
        bool,
        Field(description="Is the overwatcher running?"),
    ] = False

    enabled: Annotated[
        bool,
        Field(description="Is the overwatcher enabled?"),
    ] = False

    observing: Annotated[
        bool,
        Field(description="Is the overwatcher observing?"),
    ] = False

    cancelling: Annotated[
        bool,
        Field(description="Are observations being cancelled?"),
    ] = False

    calibrating: Annotated[
        bool,
        Field(description="Is the overwatcher taking cals?"),
    ] = False

    focusing: Annotated[
        bool,
        Field(description="Is the overwatcher focusing the telescopes?"),
    ] = False

    troubleshooting: Annotated[
        bool,
        Field(description="Is the overwatcher in troubleshooting mode?"),
    ] = False

    night: Annotated[
        bool | None,
        Field(description="Is currently night (after twilight)?"),
    ] = None

    safe: Annotated[
        bool | None,
        Field(description="Is it safe to observe?"),
    ] = None

    allow_calibrations: Annotated[
        bool,
        Field(description="Is the overwatcher allowed to take cals?"),
    ] = False

    running_calibration: Annotated[
        str | None,
        Field(description="Name of the currently running calibration"),
    ] = None

    tile_id: Annotated[
        int | None,
        Field(description="Tile ID of the current observation"),
    ] = None

    dither_position: Annotated[
        int | None,
        Field(description="Dither position of the current observation"),
    ] = None

    stage: Annotated[
        str | None,
        Field(description="Stage of the current observation"),
    ] = None

    standard_no: Annotated[
        int | None,
        Field(description="Standard number of the current observation"),
    ] = None


class OverwatcherCalibrationModel(BaseModel):
    """Overwatcher calibration model."""

    name: Annotated[
        str,
        Field(description="The name of the calibration"),
    ]

    start_time: Annotated[
        str | None,
        Field(description="The start time of the calibration"),
    ]

    max_start_time: Annotated[
        str | None,
        Field(description="The maximum start time of the calibration"),
    ]

    after: Annotated[
        str | None,
        Field(description="Start the calibration only after this one is complete"),
    ]

    time_to_cal: Annotated[
        float | None,
        Field(description="Number of seconds to the calibration start"),
    ]

    status: Annotated[
        str,
        Field(description="The status of the calibration"),
    ]

    requires_dome: Annotated[
        Literal["open", "closed"] | None,
        Field(description="Required position of the dome"),
    ]

    close_dome_after: Annotated[
        bool,
        Field(description="Close the dome after the calibration"),
    ]


router = APIRouter(prefix="/overwatcher", tags=["overwatcher"])


@router.get("/", summary="Overwatcher route")
async def get_overwatcher():
    """Not implemented."""

    return {}


@router.get(
    "/status",
    summary="Returns the status of the overwatcher",
    response_model=OverwatcherStatusModel,
)
async def get_overwatcher_status() -> OverwatcherStatusModel:
    """Returns the status of the overwatcher."""

    status: dict[str, bool]
    async with CluClient() as clu:
        status_cmd = await clu.send_command("lvm.overwatcher", "status")

        try:
            status = status_cmd.replies.get("status")
        except KeyError:
            status = {"running": False}

        observer_command = await clu.send_command("lvm.overwatcher", "observer status")

        try:
            observer_status = observer_command.replies.get("observer_status")
            observer_status.pop("observing")  # Duplicated.
        except KeyError:
            observer_status = {}

    running_cal = await route_get_calibrations_current()

    return OverwatcherStatusModel(
        **status,
        **observer_status,
        running_calibration=running_cal,
    )


@router.get("/status/enabled", summary="Is the overwatcher enabled?")
async def get_overwatcher_enabled() -> bool:
    """Returns whether the overwatcher is enabled."""

    status = await get_overwatcher_status()

    return status.enabled


@router.put(
    "/status/{enable_or_disable}",
    summary="Enable or disable the overwatcher",
    dependencies=[AuthDependency],
)
async def put_overwatcher_enabled(
    enable_or_disable: Annotated[
        Literal["enable", "disable"],
        Path(description="Whether to enable or disable the overwatcher"),
    ],
    now: Annotated[
        bool,
        Query(description="Whether to stop observing immediately"),
    ] = False,
):
    """Enables or disables the overwatcher."""

    async with CluClient() as clu:
        await clu.send_command(
            "lvm.overwatcher",
            enable_or_disable,
            "--now" if enable_or_disable == "disable" and now else "",
        )


@router.get("/status/allow_calibrations", summary="Allow calibrations?")
async def get_allow_calibrations() -> bool:
    """Returns whether the overwatcher can take calibrations."""

    status = await get_overwatcher_status()

    return status.allow_calibrations


@router.put(
    "/status/allow_calibrations/{enable_or_disable}",
    summary="Enable or disable calibrations",
    dependencies=[AuthDependency],
)
async def put_allow_calibrations_enabled(
    enable_or_disable: Annotated[
        Literal["enable", "disable"],
        Path(description="Whether to enable or disable calibrations"),
    ],
):
    """Enables or disables calibrations."""

    command: str = "calibrations "
    if enable_or_disable == "enable":
        command += "enable-calibrations"
    else:
        command += "disable-calibrations"

    async with CluClient() as clu:
        await clu.send_command("lvm.overwatcher", command)


@router.get("/logs", summary="Returns a list of log files")
async def get_logs_files_route():
    """Returns a list of log files."""

    files = pathlib.Path("/data/logs/lvmgort/").glob("[0-9]*.log")

    return sorted([file.name for file in files])


@router.get("/logs/{logfile}", summary="Returns a logfile text")
async def get_logs_data_route(
    logfile: str,
    n_lines: int | None = Query(None, description="Returns only the last N lines"),
    as_task: bool = Query(
        False,
        description="Whether to schedule this as a task.",
    ),
):
    """Returns a logfile text."""

    task = await get_gort_log_task.kiq(logfile, n_lines=n_lines)
    if as_task:
        return task.task_id

    result = await task.wait_result()
    return result.return_value


@router.get("/calibrations/list", summary="Returns the list of calibrations")
async def route_get_calibrations_list() -> list[OverwatcherCalibrationModel]:
    """Returns the list of calibrations."""

    async with CluClient() as clu:
        calibrations_cmd = await clu.send_command(
            "lvm.overwatcher",
            "calibrations list",
        )

    if calibrations_cmd.status.did_fail:
        return []

    calibrations = calibrations_cmd.replies.get("calibrations")

    return [OverwatcherCalibrationModel(**calibration) for calibration in calibrations]


@router.get("/calibrations/running", summary="Currently running calibration")
async def route_get_calibrations_current() -> str | None:
    """Returns the name of the currently running calibration, if any."""

    cals = await route_get_calibrations_list()
    for cal in cals:
        if cal.status in ["running", "retrying"]:
            return cal.name



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-07-24
# @Filename: tasks.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import pathlib

from typing import Literal

from lvmapi.app import broker
from lvmapi.tools.gort import get_gort_client


@broker.task()
async def move_dome_task(direction: Literal["open", "close"], force: bool = False):
    """Opens/closes the dome.

    Uses GORT to ensure that the telescopes are parked before moving the dome.

    """

    async with get_gort_client() as gort:
        if direction == "open":
            await gort.enclosure.open()
        else:
            await gort.enclosure.close(force=force)

    return True


@broker.task()
async def shutdown_task(disable_overwatcher: bool = False):
    """Shuts down the system."""

    async with get_gort_client() as gort:
        await gort.emergency_shutdown()

    return True


@broker.task()
async def cleanup_task(readout: bool = True):
    """Shuts down the system."""

    async with get_gort_client() as gort:
        await gort.cleanup(readout=readout)

    return True


@broker.task()
async def restart_kubernetes_deployment_task(deployment: str, confirm: bool = True):
    """Restarts a Kubernetes deployment."""

    from lvmapi.app import app

    await app.state.kubernetes.restart_deployment(deployment)

    if confirm:
        for _ in range(15):
            if deployment in app.state.kubernetes.list_deployments():
                return True
            await asyncio.sleep(1)

    else:
        return True

    raise TimeoutError(f"Timed out waiting for {deployment} to start.")


@broker.task()
async def get_exposure_data_task(mjd: int):
    """Returns the list of exposures for a given MJD."""

    from lvmapi.tools.logs import get_exposure_data

    exposure_data = get_exposure_data(mjd)

    return exposure_data


@broker.task()
async def get_gort_log_task(logfile: str, n_lines: int | None = None):
    """Returns the log for a given MJD."""

    if not logfile.endswith(".log"):
        logfile += ".log"

    PATH = pathlib.Path("/data/logs/lvmgort/")

    if not (PATH / logfile).exists():
        return ""

    with open(PATH / logfile, "r") as file:
        if n_lines is None:
            return file.read()

        data = file.read().splitlines()[-n_lines:]
        return "\n".join(data)


@broker.task()
async def park_telescopes_task():
    """Parks all telescopes."""

    async with get_gort_client() as gort:
        await gort.telescopes.home()
        await gort.telescopes.park(disable=True)

    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-07-10
# @Filename: alerts.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import warnings

import polars
from fastapi import APIRouter, Request
from pydantic import BaseModel

from lvmapi.tools.alerts import enclosure_alerts, spec_temperature_alerts
from lvmapi.tools.weather import get_weather_data, is_measurament_safe


class AlertsSummary(BaseModel):
    """Summary of alerts."""

    humidity_alert: bool | None = None
    dew_point_alert: bool | None = None
    wind_alert: bool | None = None
    rain: bool | None = None
    door_alert: bool | None = None
    camera_temperature_alert: bool | None = None
    camera_alerts: dict[str, bool] | None = None
    o2_alert: bool | None = None
    o2_room_alerts: dict[str, bool] | None = None
    heater_alert: bool | None = None
    heater_camera_alerts: dict[str, bool] | None = None


router = APIRouter(prefix="/alerts", tags=["alerts"])


@router.get("")
@router.get("/")
@router.get("/summary")
async def summary(request: Request) -> AlertsSummary:
    """Summary of alerts."""

    tasks: list[asyncio.Task] = []
    tasks.append(asyncio.create_task(spec_temperature_alerts()))
    tasks.append(asyncio.create_task(enclosure_alerts()))
    tasks.append(asyncio.create_task(get_weather_data(start_time=3600)))

    results = await asyncio.gather(*tasks, return_exceptions=True)

    camera_alerts = results[0]
    if isinstance(camera_alerts, BaseException):
        warnings.warn(f"Error getting temperature alerts: {camera_alerts}")
        camera_alerts = None
        camera_temperature_alert = None
    else:
        camera_temperature_alert = any(camera_alerts.values())

    enclosure_alerts_response = results[1]
    if isinstance(enclosure_alerts_response, BaseException):
        enclosure_alerts_response = {}

    weather_data: polars.DataFrame | BaseException = results[2]

    wind_alert = None
    humidity_alert = None
    dew_point_alert = None

    if not isinstance(weather_data, BaseException) and weather_data.height > 0:
        wind_alert = not is_measurament_safe(
            weather_data,
            "wind_speed_avg",
            threshold=35,
            reopen_value=30,
        )

        humidity_alert = not is_measurament_safe(
            weather_data,
            "relative_humidity",
            threshold=80,
            reopen_value=70,
        )

        last_weather = weather_data[-1]
        dew_point_alert = last_weather["dew_point"][0] > last_weather["temperature"][0]

    o2_alerts = {
        key: value for key, value in enclosure_alerts_response.items() if "o2_" in key
    }

    o2_alert = any(o2_alerts.values())
    rain_sensor_alarm = enclosure_alerts_response.get("rain_sensor_alarm", None)
    door_alert = enclosure_alerts_response.get("door_alert", None)

    # These fake states are just for testing.
    if request.app.state.use_fake_states:
        humidity_alert = request.app.state.fake_states["humidity_alert"]
        wind_alert = request.app.state.fake_states["wind_alert"]
        rain_sensor_alarm = request.app.state.fake_states["rain_alert"]
        door_alert = request.app.state.fake_states["door_alert"]

    return AlertsSummary(
        humidity_alert=humidity_alert,
        dew_point_alert=dew_point_alert,
        rain=rain_sensor_alarm,
        door_alert=door_alert,
        wind_alert=wind_alert,
        camera_temperature_alert=camera_temperature_alert,
        camera_alerts=camera_alerts,
        heater_alert=False,
        heater_camera_alerts={},
        o2_alert=o2_alert,
        o2_room_alerts=o2_alerts,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-11
# @Filename: telescopes.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from typing import get_args

from fastapi import APIRouter, HTTPException
from pydantic import BaseModel, Field

from sdsstools.utils import GatheringTaskGroup

from lvmapi.auth import AuthDependency
from lvmapi.tasks import park_telescopes_task
from lvmapi.tools import CluClient
from lvmapi.tools.telescopes import get_telescope_status, is_telescope_parked
from lvmapi.types import Coordinates, Telescopes


class PointingResponse(BaseModel):
    """Response model for telescope pointing."""

    ra: float = Field(description="Right Ascension in degrees")
    dec: float = Field(description="Declination in degrees")
    alt: float = Field(description="Altitude in degrees")
    az: float = Field(description="Azimuth in degrees")


class TelescopeStatusResponse(BaseModel):
    """Response model for telescope status."""

    reachable: bool = Field(
        description="Whether the telescope is reachable",
    )
    ra: float | None = Field(
        default=None,
        description="Right Ascension in degrees",
    )
    dec: float | None = Field(
        default=None,
        description="Declination in degrees",
    )
    alt: float | None = Field(
        default=None,
        description="Altitude in degrees",
    )
    az: float | None = Field(
        default=None,
        description="Azimuth in degrees",
    )
    is_tracking: bool | None = Field(
        default=None,
        description="Whether the telescope is tracking",
    )
    is_connected: bool | None = Field(
        default=None,
        description="Whether the telescope is connected",
    )
    is_slewing: bool | None = Field(
        default=None,
        description="Whether the telescope is slewing",
    )
    is_enabled: bool | None = Field(
        default=None,
        description="Whether the telescope is enabled",
    )
    is_parked: bool | None = Field(
        default=None,
        description="Whether the telescope is parked",
    )


router = APIRouter(prefix="/telescopes", tags=["telescopes"])


@router.get("/", summary="List of telescopes")
async def route_get_telescopes() -> list[str]:
    """Returns the list of telescopes."""

    return list(get_args(Telescopes))


@router.get("/coordinates", summary="Pointing of al ltelescopes")
async def route_get_coordinates() -> dict[str, PointingResponse]:
    """Gets the pointings of all telescopes."""

    telescopes = get_args(Telescopes)
    pointings = await asyncio.gather(*[route_get_pointing(tel) for tel in telescopes])

    return {tel: pointing for tel, pointing in zip(telescopes, pointings)}


@router.get(
    "/{telescope}/pointing",
    summary="Telescope pointing",
    response_model=PointingResponse,
)
async def route_get_pointing(telescope: Telescopes) -> PointingResponse:
    """Gets the pointing of a telescope."""

    try:
        async with CluClient() as client:
            status_cmd = await client.send_command(f"lvm.{telescope}.pwi", "status")

        ra = status_cmd.replies.get("ra_apparent_hours") * 15
        dec = status_cmd.replies.get("dec_apparent_degs")
        alt = status_cmd.replies.get("altitude_degs")
        az = status_cmd.replies.get("azimuth_degs")

    except Exception:
        raise HTTPException(
            status_code=500,
            detail="Error retrieving telescope information.",
        )

    return PointingResponse(ra=ra, dec=dec, alt=alt, az=az)


@router.get("/{telescope}/{coordinate}", summary="Telescope coordinates")
async def route_get_tel_coord(telescope: Telescopes, coordinate: Coordinates) -> float:
    """Returns a given coordinate for a telescope."""

    pointing = await route_get_pointing(telescope)
    return getattr(pointing, coordinate)


@router.get("/status", summary="Telescope status")
async def route_get_telescope_status():
    """Gets the status of all telescopes."""

    status = await asyncio.gather(
        *[get_telescope_status(tel) for tel in get_args(Telescopes)],
        return_exceptions=True,
    )

    response: dict[Telescopes, TelescopeStatusResponse] = {}
    for tel, stat in zip(get_args(Telescopes), status):
        if isinstance(stat, BaseException):
            response[tel] = TelescopeStatusResponse(reachable=False)
        else:
            response[tel] = TelescopeStatusResponse(reachable=True, **stat)

        is_parked = is_telescope_parked(response[tel].model_dump())
        response[tel].is_parked = is_parked

    return response


@router.get("/parked", summary="Are telescopes parked?")
async def route_get_parked() -> dict[Telescopes, bool | None]:
    """Returns whether the telescopes are parked or not."""

    status = await route_get_telescope_status()

    response: dict[Telescopes, bool | None] = {}
    for telescope, status in status.items():
        response[telescope] = is_telescope_parked(status.model_dump())

    return response


@router.get("/park", summary="Park all telescopes", dependencies=[AuthDependency])
async def route_park_telescopes() -> str:
    """Parks all telescopes. Scheduled as a task."""

    task = await park_telescopes_task.kiq()

    return task.task_id


@router.get("/connect", summary="Connect all telescopes", dependencies=[AuthDependency])
async def route_connect_telescopes() -> bool:
    """Connects all telescopes."""

    async with CluClient() as client:
        async with GatheringTaskGroup() as gr:
            for tel in get_args(Telescopes):
                gr.create_task(client.send_command(f"lvm.{tel}.pwi", "setConnected 1"))

    for cmd in gr.results():
        if cmd.status.did_fail:
            return False

    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-07-24
# @Filename: tools.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import base64
import json
import pathlib
from datetime import UTC, datetime

from typing import Annotated, get_args, overload

import polars
import psycopg
import psycopg.rows
import psycopg.sql
from pydantic import BaseModel, Field

from lvmopstools.devices.ion import read_ion_pumps, toggle_ion_pump
from lvmopstools.devices.specs import (
    exposure_etr,
    spectrograph_mechanics,
    spectrograph_pressures,
    spectrograph_status,
    spectrograph_temperature_label,
    spectrograph_temperatures,
)
from lvmopstools.devices.thermistors import channel_to_valve, read_thermistors

from lvmapi import config
from lvmapi.tools.general import get_db_connection
from lvmapi.tools.influxdb import query_influxdb
from lvmapi.types import Cameras, Sensors, SpecStatus, Spectrographs


__all__ = [
    "spectrograph_temperature_label",
    "spectrograph_temperatures",
    "spectrograph_temperatures_history",
    "spectrograph_pressures",
    "spectrograph_mechanics",
    "spectrograph_status",
    "read_thermistors",
    "read_thermistors_influxdb",
    "read_ion_pumps",
    "toggle_ion_pump",
    "exposure_etr",
    "retrieve_fill_measurements",
    "register_ln2_fill",
    "retrieve_fill_metadata",
    "FillMetadataModel",
    "FillMetadataReturn",
    "SpecToStatus",
    "get_fill_list",
]


SpecToStatus = dict[Spectrographs, SpecStatus]


async def spectrograph_temperatures_history(
    start: str = "-30m",
    stop: str = "now()",
    camera: str | None = None,
    sensor: str | None = None,
):
    time_range = f"|> range(start: {start}, stop: {stop})"

    spec_filter = r'|> filter(fn: (r) => r["_measurement"] =~ /lvmscp\.sp[1-3]/)'
    if camera is not None:
        spec_filter = (
            f'|> filter(fn: (r) => r["_measurement"] == "lvmscp.sp{camera[-1]}")'
        )

    sensor_filter = r'|> filter(fn: (r) => r["_field"] =~ /mod(2|12)\/temp[abc]/)'
    if camera is not None and sensor is not None:
        field = spectrograph_temperature_label(camera[0], sensor)
        sensor_filter = f'|> filter(fn: (r) => r["_field"] == "status.{field}")'

    query = rf"""
    from(bucket: "actors")
        {time_range}
        {spec_filter}
        {sensor_filter}
    """

    results = await query_influxdb(query)

    if len(results) == 0:
        return polars.DataFrame(
            None,
            schema={
                "time": polars.String,
                "camera": polars.String,
                "sensor": polars.String,
                "temperature": polars.Float32,
            },
        )

    final_df = results.select(
        time=polars.col._time,
        camera=polars.lit(None, dtype=polars.String),
        sensor=polars.lit(None, dtype=polars.String),
        temperature=polars.col._value.cast(polars.Float32),
    )

    for spec in ["sp1", "sp2", "sp3"]:
        for cc in get_args(Cameras):
            for ss in get_args(Sensors):
                label = spectrograph_temperature_label(cc, ss)

                _measurement = f"lvmscp.{spec}"
                _field = f"status.{label}"

                final_df[
                    (
                        (results["_measurement"] == _measurement)
                        & (results["_field"] == _field)
                    ).arg_true(),
                    "camera",
                ] = f"{cc}{spec[-1]}"

                final_df[(results["_field"] == _field).arg_true(), "sensor"] = ss

    final_df = final_df.select(polars.col(["time", "camera", "sensor", "temperature"]))
    final_df = final_df.sort(["time", "camera", "sensor"])

    if camera:
        final_df = final_df.filter(polars.col.camera == camera)

    if sensor:
        final_df = final_df.filter(polars.col.sensor == sensor)

    return final_df


@overload
async def read_thermistors_influxdb(interval: float) -> polars.DataFrame: ...


@overload
async def read_thermistors_influxdb(interval: None) -> dict[str, bool]: ...


async def read_thermistors_influxdb(
    interval: float | None = None,
) -> polars.DataFrame | dict[str, bool]:
    """Returns thermistor states from InfluxDB.

    Parameters
    ----------
    interval
        How far to look back in seconds. If None, returns the latest value.

    Returns
    -------
    states
        If ``interval=None``, a dictionary of thermistor states. Otherwise a
        Polars dataframe with the thermistor states, one row per measurement
        within the interval.

    """

    if interval is None:
        start = "-1m"
        last = "|> last()"
    else:
        start = f"-{interval}s"
        last = ""

    query = f"""
from(bucket: "spec")
  |> range(start: {start}, stop: now())
  |> filter(fn: (r) => r["_measurement"] == "thermistors")
  |> filter(fn: (r) => r["channel_name"] != "")
  {last}
  |> group(columns: ["_field"])
  |> keep(columns: ["_time", "channel_name", "_value"])
"""

    data = await query_influxdb(query)

    if interval is None:
        result: dict[str, bool] = {}
        for row in data.iter_rows(named=True):
            result[row["channel_name"]] = bool(row["_value"])
        return result

    df = data.select(
        time=polars.col._time,
        channel=polars.col.channel_name,
        state=polars.col._value.cast(polars.Boolean),
    )
    df = df.sort(["channel", "time"])

    return df


async def retrieve_fill_measurements(
    start: float | datetime,
    end: float | datetime,
) -> polars.DataFrame:
    """Retrieves LN2 fill data from InfluxDB.

    This function retrieves the LN2 fill data from InfluxDB for a given interval.
    That includes LN2 and CCD temperatures, pressures, NPS outlet status, and
    thermistor status. This query is intended mostly for ``lvmcryo`` to collect
    metrology data after a fill.

    Parameters
    ----------
    start
        The start of the interval. Either a ``datetime`` object in UTC or a Unix
        timestamp.
    end
        The end of the interval.

    Returns
    -------
    fill_data
        A Polars DataFrame with the fill data.

    """

    query_start_time: int | str
    query_end_time: int | str

    if isinstance(start, datetime):
        query_start_time = start.strftime("%Y-%m-%dT%H:%M:%SZ")
    else:
        query_start_time = int(start)

    if isinstance(end, datetime):
        query_end_time = end.strftime("%Y-%m-%dT%H:%M:%SZ")
    else:
        query_end_time = int(end)

    # Get data from the "spec" bucket (pressure and thermistors).

    query_cryo = f"""
        from(bucket: "spec")
        |> range(start: {query_start_time}, stop: {query_end_time})
        |> filter(fn: (r) => r["_measurement"] == "pressure" or r["_measurement"] == "thermistors")
        |> filter(fn: (r) => (r["_measurement"] == "thermistors" and r["_field"] =~ /channel[0-9]+/) or
                             (r["_measurement"] == "pressure" and r["_field"] == "cmb"))
        |> aggregateWindow(every: 1s, fn: mean, createEmpty: false)
        |> yield(name: "mean")
    """  # noqa: E501

    data_cryo = await query_influxdb(query_cryo)

    pressure_data = data_cryo.filter(polars.col._measurement == "pressure")

    # Select pressures and pivot. Pressure is the data we poll the least so for
    # short intervals there may not be any data.
    if len(pressure_data) > 0:
        pressures = (
            pressure_data.select(
                time=polars.col._time.cast(polars.Datetime("ms", UTC)),
                ccd=polars.col.ccd,
                value=polars.col._value,
            )
            .with_columns(ccd="pressure_" + polars.col.ccd)
            .pivot("ccd", index="time", values="value")
            .sort("time")
        )
    else:
        pressures = polars.DataFrame(None, schema={"time": polars.Datetime("ms", UTC)})

    # Select thermistors and pivot.
    ch_to_valve = channel_to_valve()
    thermistors = (
        data_cryo.filter(polars.col._measurement == "thermistors")
        .select(
            time=polars.col._time.cast(polars.Datetime("ms", UTC)),
            field=polars.col._field,
            value=polars.col._value.cast(polars.Boolean),
        )
        .with_columns(
            channel="thermistor_"
            + polars.col.field.str.replace("channel", "")
            .cast(polars.Int32)
            .replace_strict(ch_to_valve, default=None)
        )
        .drop("field")
        .drop_nulls()
        .pivot("channel", index="time", values="value")
        .sort("time")
    )

    # Get temperatures and pivot
    data_temp = await spectrograph_temperatures_history(
        start=str(query_start_time),
        stop=str(query_end_time),
    )

    temp_df = (
        data_temp.select(
            time=polars.col.time.cast(polars.Datetime("ms", UTC)),
            sensor="temp_" + polars.col.camera + "_" + polars.col.sensor,
            value=polars.col.temperature,
        )
        .pivot(
            "sensor",
            index="time",
            values="value",
        )
        .sort("time")
    )

    # Create a uniform range of times.
    if not isinstance(start, datetime):
        start = datetime.fromtimestamp(start, UTC)
    if not isinstance(end, datetime):
        end = datetime.fromtimestamp(end, UTC)

    time_range = (
        polars.datetime_range(start, end, "1s", eager=True, time_unit="ms")
        .alias("time")
        .to_frame()
    )

    # Join with the data frames.
    data = (
        time_range.join_asof(pressures, on="time", strategy="nearest")
        .join_asof(thermistors, on="time", strategy="nearest")
        .join_asof(temp_df, on="time", strategy="nearest")
        .with_columns(polars.all().forward_fill().backward_fill())
    )

    return data


async def register_ln2_fill(
    *,
    pk: int | None,
    action: str,
    start_time: str | None,
    end_time: str | None,
    purge_start: str | None,
    purge_complete: str | None,
    fill_start: str | None,
    fill_complete: str | None,
    fail_time: str | None,
    abort_time: str | None,
    failed: bool,
    aborted: bool,
    error: str | None,
    log_file: str | None,
    json_file: str | None,
    configuration: dict | None,
    log_data: list[dict] | None,
    plot_paths: dict[str, str] | None,
    valve_times: dict[str, dict[str, str | bool | None]] | None,
    complete: bool = True,
) -> int:
    """Registers LN2 fill data in the database."""

    table: list[str] = config["database.tables.ln2_fill"].split(".")

    columns = """(start_time, end_time, purge_start, purge_complete,
fill_start, fill_complete, fail_time, abort_time,
failed, aborted, error, action, log_file, json_file,
configuration, log_data, plot_paths, valve_times, complete)"""

    values_placeholders = """(%(start_time)s, %(end_time)s, %(purge_start)s,
%(purge_complete)s, %(fill_start)s, %(fill_complete)s, %(fail_time)s, %(abort_time)s,
%(failed)s, %(aborted)s, %(error)s, %(action)s, %(log_file)s, %(json_file)s,
%(configuration)s, %(log_data)s, %(plot_paths)s, %(valve_times)s, %(complete)s)"""

    params = {
        "start_time": start_time,
        "end_time": end_time,
        "purge_start": purge_start,
        "purge_complete": purge_complete,
        "fill_start": fill_start,
        "fill_complete": fill_complete,
        "fail_time": fail_time,
        "abort_time": abort_time,
        "failed": failed,
        "aborted": aborted,
        "error": error,
        "action": action,
        "log_file": log_file,
        "json_file": json_file,
        "configuration": json.dumps(configuration) if configuration else None,
        "log_data": json.dumps(log_data) if log_data else None,
        "plot_paths": json.dumps(plot_paths) if plot_paths else None,
        "valve_times": json.dumps(valve_times) if valve_times else None,
        "complete": complete,
    }

    if pk is None:
        query = psycopg.sql.SQL("INSERT INTO {} {} VALUES {} RETURNING pk;").format(
            psycopg.sql.Identifier(*table),
            psycopg.sql.SQL(columns),
            psycopg.sql.SQL(values_placeholders),
        )
    else:
        query = psycopg.sql.SQL(
            "UPDATE {} SET {} = {} WHERE pk = %(pk)s RETURNING pk;"
        ).format(
            psycopg.sql.Identifier(*table),
            psycopg.sql.SQL(columns),
            psycopg.sql.SQL(values_placeholders),
        )
        params["pk"] = pk

    async with await get_db_connection() as aconn:
        async with aconn.cursor() as acur:
            await acur.execute(query, params)
            await aconn.commit()
            returned = await acur.fetchone()

    if returned is None:
        raise ValueError("Could not retrieve primary key.")

    return returned[0]


class FillMetadataModel(BaseModel):
    """Request model for the ``/fills/register`` endpoint."""

    action: str
    complete: bool
    start_time: datetime | None
    end_time: datetime | None
    purge_start: datetime | None
    purge_complete: datetime | None
    fill_start: datetime | None
    fill_complete: datetime | None
    fail_time: datetime | None
    abort_time: datetime | None
    failed: bool
    aborted: bool
    error: str | None
    log_file: str | None
    json_file: str | None
    configuration: dict | None
    log_data: list[dict] | None
    plot_paths: dict | None
    valve_times: dict | None


class FillMetadataReturn(BaseModel):
    """Return model for `.retrieve_fill_metadata`."""

    pk: Annotated[
        int,
        Field(description="Primary key of the LN2 fill record"),
    ]
    action: Annotated[
        str,
        Field(description="LN2 action performed"),
    ]
    complete: Annotated[
        bool,
        Field(description="Is the fill complete?"),
    ]
    start_time: Annotated[
        datetime | None,
        Field(description="Start time"),
    ]
    end_time: Annotated[
        datetime | None,
        Field(description="End time"),
    ]
    purge_start: Annotated[
        datetime | None,
        Field(description="Purge start time"),
    ]
    purge_complete: Annotated[
        datetime | None,
        Field(description="Purge finish time"),
    ]
    fill_start: Annotated[
        datetime | None,
        Field(description="Fill start time"),
    ]
    fill_complete: Annotated[
        datetime | None,
        Field(description="Fill finish time"),
    ]
    fail_time: Annotated[
        datetime | None,
        Field(description="Time of failure"),
    ]
    abort_time: Annotated[
        datetime | None,
        Field(description="Time of abort"),
    ]
    failed: Annotated[
        bool,
        Field(description="Did the action fail?"),
    ]
    aborted: Annotated[
        bool,
        Field(description="Was the action aborted?"),
    ]
    error: Annotated[
        str | None,
        Field(description="Error message"),
    ]
    log_data: Annotated[
        str | None,
        Field(description="Reconstructed log file"),
    ]
    configuration: Annotated[
        dict | None,
        Field(description="Configuration data"),
    ]
    plot_data: Annotated[
        dict[str, str | bytes],
        Field(description="Plot images encoded as base64"),
    ] = {}
    valve_times: Annotated[
        dict[str, dict[str, str | bool | None]] | None,
        Field(description="Valve open/close times"),
    ]


async def retrieve_fill_metadata(
    pk: int,
    transparent_plots: bool = False,
    as_base64: bool = False,
):
    """Retrieves metadata for a given LN2 fill from the database."""

    uri = config["database.uri"]
    table: list[str] = config["database.tables.ln2_fill"].split(".")

    async with await psycopg.AsyncConnection.connect(uri) as aconn:
        async with aconn.cursor(row_factory=psycopg.rows.dict_row) as acur:
            await acur.execute(
                psycopg.sql.SQL("SELECT * FROM {} WHERE pk = (%s)").format(
                    psycopg.sql.Identifier(*table)
                ),
                (pk,),
            )
            db_data = await acur.fetchone()

    if db_data is None:
        raise ValueError(f"Cannot find LN2 fill with pk={pk}.")

    data = FillMetadataModel(**db_data)

    # Reconstruct log from JSON.
    log_data: str | None = None
    if data.log_data:
        lines = []
        for line in data.log_data:
            timestamp = datetime.fromisoformat(line["timestamp"])
            log_ts = timestamp.strftime("%Y-%m-%d %H:%M:%S")
            log_ts += f".{timestamp.microsecond}"[0:2]
            level = line["level"]
            message = line["message"]
            lines.append(f"{log_ts} - {level} - {message}")
        log_data = "\n".join(lines)

    # Encode plot images as base64 bytes.
    plot_data: dict[str, str | bytes] = {}
    if data.plot_paths:
        for key, value in data.plot_paths.items():
            if (transparent_plots and value.endswith("_transparent.png")) or (
                not transparent_plots
                and value.endswith(".png")
                and "transparent" not in value
            ):
                path = pathlib.Path(value)
                if not path.exists():
                    continue

                if as_base64:
                    with open(path, "rb") as f:
                        plot_data[key] = base64.b64encode(f.read())
                else:
                    plot_data[key] = str(path.absolute())

    return FillMetadataReturn(
        pk=pk,
        action=data.action,
        complete=data.complete,
        start_time=data.start_time,
        end_time=data.end_time,
        purge_start=data.purge_start,
        purge_complete=data.purge_complete,
        fill_start=data.fill_start,
        fill_complete=data.fill_complete,
        fail_time=data.fail_time,
        abort_time=data.abort_time,
        failed=data.failed,
        aborted=data.aborted,
        error=data.error,
        log_data=log_data,
        configuration=data.configuration,
        plot_data=plot_data,
        valve_times=data.valve_times,
    )


async def get_fill_list():
    """Returns a mapping of LN2 fill PK to start time."""

    uri = config["database.uri"]
    table: list[str] = config["database.tables.ln2_fill"].split(".")

    async with await psycopg.AsyncConnection.connect(uri) as aconn:
        async with aconn.cursor() as acur:
            await acur.execute(
                psycopg.sql.SQL("SELECT pk, start_time FROM {}").format(
                    psycopg.sql.Identifier(*table)
                )
            )
            db_data = await acur.fetchall()

    result: dict[int, str] = {}
    for row in db_data:
        result[row[0]] = row[1].strftime("%Y-%m-%d %H:%M:%S")

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-11
# @Filename: types.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import Literal


Spectrographs = Literal["sp1", "sp2", "sp3"]
SpecStatus = Literal["idle", "exposing", "reading", "error", "unknown"]
Cameras = Literal["r", "z", "b"]
CamSpec = Literal["r1", "z1", "b1", "r2", "z2", "b2", "r3", "z3", "b3"]
Sensors = Literal["ccd", "ln2"]

Telescopes = Literal["sci", "spec", "skye", "skyw"]
Frames = Literal["radec", "altaz"]
Coordinates = Literal["ra", "dec", "alt", "az"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-03-24
# @Filename: schedule.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import pathlib

import astroplan
import numpy
import polars
from astropy import units as uu
from astropy.time import Time
from cachetools import TTLCache, cached

from sdsstools import get_sjd


EPHEMERIS_FILE = pathlib.Path(__file__).parent / "../data/ephemeris.parquet"


@cached(TTLCache(maxsize=3, ttl=3600))
def get_ephemeris_data(filename: pathlib.Path | str) -> polars.DataFrame:
    """Returns and caches the data from the ephemeris file."""

    return polars.read_parquet(filename)


def sjd_ephemeris(sjd: int, twilight_horizon: float = -15) -> polars.DataFrame:
    """Returns the ephemeris for a given SJD."""

    observer = astroplan.Observer.at_site("Las Campanas Observatory")
    observer.pressure = 0.76 * uu.bar

    # Calculate Elevation of True Horizon. Astroplan does not provide this directly.
    # See https://github.com/astropy/astroplan/issues/242
    h_observer = observer.elevation
    R_earth = 6378100.0 * uu.m
    dd = numpy.sqrt(h_observer * (2 * R_earth + h_observer))
    phi = (numpy.arccos((dd / R_earth).value) * uu.radian).to(uu.deg)
    hzel = phi - 90 * uu.deg

    # Calculate time at ~15UT, which corresponds to about noon at LCO, so always
    # before the beginning of the night.
    time = Time(sjd - 0.35, format="mjd", scale="utc")

    sunset = observer.sun_set_time(
        time,
        which="next",
        horizon=hzel - 0.5 * uu.deg,  # Apparent size of the Sun.
    )
    sunset_twilight = observer.sun_set_time(
        time,
        which="next",
        horizon=twilight_horizon * uu.deg,
    )

    sunrise = observer.sun_rise_time(
        time,
        which="next",
        horizon=hzel - 0.5 * uu.deg,
    )
    sunrise_twilight = observer.sun_rise_time(
        time,
        which="next",
        horizon=twilight_horizon * uu.deg,
    )

    moon_illumination = observer.moon_illumination(time)

    df = polars.DataFrame(
        [
            (
                sjd,
                time.isot.split("T")[0],
                sunset.jd,
                sunset_twilight.jd,
                sunrise_twilight.jd,
                sunrise.jd,
                moon_illumination,
            )
        ],
        schema={
            "SJD": polars.Int32,
            "date": polars.String,
            "sunset": polars.Float64,
            "twilight_end": polars.Float64,
            "twilight_start": polars.Float64,
            "sunrise": polars.Float64,
            "moon_illumination": polars.Float32,
        },
        orient="row",
    )

    return df


def create_schedule(
    start_sjd: int,
    end_sjd: int,
    twilight_horizon: float = -15,
) -> polars.DataFrame:
    """Creates a schedule for the given time range.

    Parameters
    ----------
    start_sjd
        The initial SJD.
    end_sjd
        The final SJD of the schedule.

    Returns
    -------
    schedule
        The schedule as a Polars dataframe.

    """

    start_sjd = start_sjd or get_sjd("LCO")

    ephemeris: list[polars.DataFrame] = []
    for sjd in range(start_sjd, end_sjd + 1):
        sjd_eph = sjd_ephemeris(sjd, twilight_horizon=twilight_horizon)
        ephemeris.append(sjd_eph)

    return polars.concat(ephemeris)


def get_ephemeris_summary(sjd: int | None = None) -> dict:
    """Returns a summary of the ephemeris for a given SJD."""

    sjd = sjd or get_sjd("LCO")

    from_file = True
    eph = get_ephemeris_data(EPHEMERIS_FILE)

    data = eph.filter(polars.col("SJD") == sjd)
    if len(data) == 0:
        data = sjd_ephemeris(sjd)
        from_file = False

    sunset = Time(data["sunset"][0], format="jd")
    sunrise = Time(data["sunrise"][0], format="jd")
    twilight_end = Time(data["twilight_end"][0], format="jd")
    twilight_start = Time(data["twilight_start"][0], format="jd")

    time_to_sunset = (sunset - Time.now()).to(uu.h).value
    time_to_sunrise = (sunrise - Time.now()).to(uu.h).value

    is_twilight_evening = (
        Time(data["sunset"][0], format="jd")
        < Time.now()
        < Time(data["twilight_end"][0], format="jd")
    )

    is_twilight_morning = (
        Time(data["twilight_start"][0], format="jd")
        < Time.now()
        < Time(data["sunrise"][0], format="jd")
    )

    is_night = (
        Time(data["twilight_end"][0], format="jd")
        < Time.now()
        < Time(data["twilight_start"][0], format="jd")
    )

    return {
        "SJD": int(sjd),
        "request_jd": float(Time.now().jd),
        "date": data["date"][0],
        "sunset": float(sunset.jd),
        "twilight_end": float(twilight_end.jd),
        "twilight_start": float(twilight_start.jd),
        "sunrise": float(sunrise.jd),
        "is_night": bool(is_night),
        "is_twilight": bool(is_twilight_evening or is_twilight_morning),
        "time_to_sunset": round(float(time_to_sunset), 3),
        "time_to_sunrise": round(float(time_to_sunrise), 3),
        "moon_illumination": round(float(data["moon_illumination"][0]), 3),
        "from_file": from_file,
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-06-29
# @Filename: kubernetes.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import datetime
import os
import pathlib
import warnings

import kubernetes
from gort.tools import is_notebook
from kubernetes.utils import create_from_yaml

from lvmapi import config


class Kubernetes:
    """Interface with the Kubernetes cluster."""

    def __init__(self):
        self.is_notebook = is_notebook()

        if os.getenv("KUBERNETES_SERVICE_HOST"):
            self.is_pod = True
        else:
            self.is_pod = False

        # If we are in a notebook, we assume it's the one running in the Jupyter
        # Lab deployment, which is configured to have access to the cluster.
        if self.is_notebook or self.is_pod:
            kubernetes.config.load_incluster_config()
        else:
            kubernetes.config.load_config()

        self.v1 = kubernetes.client.CoreV1Api()
        self.apps_v1 = kubernetes.client.AppsV1Api()

    def list_namespaces(self):
        """Returns a list of namespaces."""

        namespace_info = self.v1.list_namespace()
        namespaces = [item.metadata.name for item in namespace_info.items]

        return namespaces

    def list_deployments(self):
        """Returns a list of deployments."""

        deployment_info = self.apps_v1.list_deployment_for_all_namespaces()
        deployments = [item.metadata.name for item in deployment_info.items]

        return deployments

    def get_deployment_info(self, deployment: str):
        """Returns the deployment info for a deployment."""

        deployment_info = self.apps_v1.list_deployment_for_all_namespaces()

        for item in deployment_info.items:
            meta = item.metadata
            if meta.name == deployment:
                return item.to_dict()

        raise ValueError(f"Deployment {deployment!r} not found.")

    def get_deployment_namespace(self, deployment: str):
        """Returns the namespace of a deployment."""

        deployment_info = self.apps_v1.list_deployment_for_all_namespaces()

        for item in deployment_info.items:
            meta = item.metadata
            if meta.name == deployment:
                return meta.namespace

        return None

    def get_yaml_file(self, name: str):
        """Finds and returns the contents of a Kubernetes YAML file."""

        path = pathlib.Path(config["kubernetes.deployments_path"])

        files = list(path.glob(f"**/{name}.y*ml"))

        if files is None or len(files) == 0:
            raise ValueError(f"No YAML file found for {name!r}.")
        elif len(files) > 1:
            raise ValueError(f"Multiple YAML files found for {name!r}.")

        return files[0]

    def apply_from_file(self, name: str | pathlib.Path):
        """Applies a YAML file.

        Parameters
        ----------
        name
            The full path to the file to apply. If the path is relative,
            the file will be searched in the directory for YAML files
            defined in the configuration.

        """

        if isinstance(name, pathlib.Path) or os.path.isabs(name):
            path = pathlib.Path(name)
        else:
            path = self.get_yaml_file(name)

        deployments = create_from_yaml(
            kubernetes.client.ApiClient(),
            yaml_file=str(path),
        )

        return [dep[0].metadata.name for dep in deployments]

    def delete_deployment(self, deployment: str):
        """Deletes resources from a YAML file.

        Parameters
        ----------
        deployment
            The deployment to delete.

        """

        namespace = self.get_deployment_namespace(deployment)
        if namespace is None:
            raise ValueError(f"Deployment {deployment!r} not found.")

        self.apps_v1.delete_namespaced_deployment(deployment, namespace)

    async def restart_deployment(self, deployment: str, from_file: bool = True):
        """Restarts a deployment.

        If the deployment is running, does a rollout restart. Otherwise looks
        for the deployment file and applies it.

        """

        if deployment in self.list_deployments() and not from_file:
            namespace = self.get_deployment_namespace(deployment)
            if namespace is None:
                raise ValueError(f"Namespace not found for deployment {deployment}.")

            # Create a patch for the current deployment saying that
            # it was restarted now, and it will.
            now = datetime.datetime.now(datetime.timezone.utc)
            now = str(now.isoformat("T") + "Z")
            body = {
                "spec": {
                    "template": {
                        "metadata": {
                            "annotations": {"kubectl.kubernetes.io/restartedAt": now}
                        }
                    }
                }
            }

            self.apps_v1.patch_namespaced_deployment(
                deployment,
                namespace,
                body,
                pretty="true",
            )

        else:
            try:
                file_ = self.get_yaml_file(deployment)
            except ValueError as err:
                raise RuntimeError(f"Failed restarting from file: {err} ")

            if deployment in self.list_deployments():
                self.delete_deployment(deployment)
                await asyncio.sleep(3)  # Give some time for the pods to exit.
            else:
                warnings.warn(f"{deployment!r} is not running.")

            self.apply_from_file(file_)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-08-06
# @Filename: log.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
from datetime import datetime

from typing import Annotated, Literal

from fastapi import APIRouter, Body, Path, Query
from pydantic import BaseModel, Field

from sdsstools import get_sjd

from lvmapi.routers.notifications import Notification
from lvmapi.tasks import get_exposure_data_task
from lvmapi.tools.logs import (
    add_night_log_comment,
    create_night_log_entry,
    delete_night_log_comment,
    email_night_log,
    get_exposure_data,
    get_exposure_table_ascii,
    get_exposures,
    get_night_log_data,
    get_night_log_mjds,
    get_night_metrics,
    get_plaintext_night_log,
    get_spectro_mjds,
)


class NightLogComment(BaseModel):
    """A comment in the night log."""

    pk: Annotated[int, Field(description="The primary key of the comment")]
    date: Annotated[datetime, Field(description="The time the comment was loaded")]
    comment: Annotated[str, Field(description="The comment text")]


class NightLogData(BaseModel):
    """The night log data for an MJD."""

    mjd: Annotated[
        int,
        Field(description="The MJD associated with the comments"),
    ]
    current: Annotated[
        bool,
        Field(description="Whether the night log is for the current MJD"),
    ]
    exists: Annotated[
        bool,
        Field(description="Whether the night log exists in the DB"),
    ]
    sent: Annotated[
        bool,
        Field(description="Whether the night log has been sent"),
    ] = False
    observers: Annotated[
        str | None,
        Field(description="The observers that took the data"),
    ] = None
    comments: Annotated[
        dict[str, list[NightLogComment]],
        Field(description="The list of comments, organised by category"),
    ] = {}
    metrics: Annotated[
        NightMetrics,
        Field(description="The night metrics"),
    ]
    exposure_table: Annotated[
        str | None,
        Field(description="The exposure table for the night log"),
    ] = None
    notifications: Annotated[
        list[Notification],
        Field(description="The list of notifications for the night log"),
    ] = []


class NightLogPostComment(BaseModel):
    """A comment to add to the night log."""

    mjd: Annotated[
        int,
        Field(description="The MJD associated with the comment"),
    ]
    category: Annotated[
        Literal["weather", "issues", "other", "observers", "overwatcher"],
        Field(description="The category of the comment"),
    ]
    comment: Annotated[
        str,
        Field(description="The comment text"),
    ]
    pk: Annotated[
        int | None,
        Field(
            description="The primary key of the comment. If provided, "
            "the comment will be updated."
        ),
    ] = None


class NightMetrics(BaseModel):
    """Model for night metrics."""

    sjd: Annotated[int, Field(description="The Sloan MJD")]
    twilight_end: Annotated[
        float,
        Field(description="Evening twilight as JD"),
    ]
    twilight_start: Annotated[
        float,
        Field(description="Morning twilight as JD"),
    ]
    night_length: Annotated[
        float,
        Field(description="Night length in seconds"),
    ]
    n_object_exps: Annotated[
        int,
        Field(description="Number of science exposures"),
    ]
    total_exp_time: Annotated[
        float,
        Field(description="Total time exposing science in seconds"),
    ]
    time_lost: Annotated[
        float,
        Field(description="Time not exposing in seconds"),
    ]
    efficiency_no_readout: Annotated[
        float,
        Field(description="Efficiency not taking into account readout, as percentage"),
    ]
    efficiency_readout: Annotated[
        float,
        Field(description="Efficiency assuming 50s readout, as percentage"),
    ]
    efficiency_nominal: Annotated[
        float,
        Field(description="Efficiency assuming 90s overhead, as percentage"),
    ]
    night_started: Annotated[
        bool,
        Field(description="Whether the night has started"),
    ]
    night_ended: Annotated[
        bool,
        Field(description="Whether the night has ended"),
    ]


router = APIRouter(prefix="/logs", tags=["logs"])


@router.get("/")
async def route_get_log():
    """Not implemented."""

    return {}


@router.get("/exposures/mjds", summary="List of MJDs with spectrograph data")
async def route_get_spectro_mjds():
    """Returns a list of MJDs with spectrograph data (or at least a folder)."""

    mjds = await asyncio.get_event_loop().run_in_executor(None, get_spectro_mjds)
    return mjds


@router.get(
    "/exposures/data/{mjd}",
    summary="Returns data from exposures for an MJD.",
)
async def route_get_exposures_data(
    mjd: Annotated[
        int,
        Path(description="The SJD (Sloan-flavoured MJD) for which to list exposures."),
    ],
    as_task: Annotated[
        bool,
        Query(description="Whether to schedule this as a task."),
    ] = False,
):
    """Returns a log of exposures for an MJD.."""

    if as_task is False:
        executor = asyncio.get_event_loop().run_in_executor
        exposure_data = await executor(None, get_exposure_data, mjd)
        return exposure_data

    task = await get_exposure_data_task.kiq(mjd)
    return task.task_id


@router.get("/exposures/{mjd}", summary="Returns a list of exposures for an MJD.")
async def route_get_exposures(
    mjd: Annotated[
        int,
        Path(description="The SJD (Sloan-flavoured MJD) for which to list exposures."),
    ],
):
    """Returns a list of exposures for an MJD."""

    executor = asyncio.get_event_loop().run_in_executor
    exposures = await executor(None, get_exposures, mjd)

    return list(map(str, exposures))


@router.get("/night-logs", summary="List of night log MJDs.")
async def route_get_night_logs():
    """Returns a list of MJDs with night log data."""

    mjds = await get_night_log_mjds()
    return mjds


@router.get("/night-logs/create", summary="Create night log entry")
async def route_get_night_logs_create():
    """Creates a night log entry for the current MJD."""

    mjd = await create_night_log_entry()
    return mjd


@router.post("/night-logs/comments/add", summary="Add night log comment")
async def route_post_night_logs_add_comment(
    data: Annotated[NightLogPostComment, Body(description="The comment to add")],
):
    """Adds a comment to a night log."""

    await add_night_log_comment(
        data.mjd,
        data.comment,
        category=data.category,
        comment_pk=data.pk,
    )


@router.get(
    "/night-logs/comments/delete/{pk}",
    summary="Delete night log comment",
)
async def route_get_night_logs_delete_comment(
    pk: Annotated[int, Path(description="The primary key of the comment")],
):
    """Deletes a comment from a night log."""

    await delete_night_log_comment(pk)


@router.get("/night-logs/{mjd}", summary="Night log data for an MJD")
async def route_get_night_logs_mjd(
    mjd: Annotated[
        int,
        Path(
            description="The MJD for which to retrieve night log data. "
            "Use 0 for tonight's log."
        ),
    ],
) -> NightLogData:
    """Returns the night log data for an MJD."""

    from .notifications import route_get_notifications

    mjd = mjd if mjd > 0 else get_sjd("LCO")
    data = await get_night_log_data(mjd)

    comments = {
        category: [NightLogComment(**comment) for comment in comments]
        for category, comments in data.pop("comments", {}).items()
    }

    exposure_table_ascii = await get_exposure_table_ascii(mjd)
    notifications = await route_get_notifications(mjd)
    metrics = get_night_metrics(mjd)

    return NightLogData(
        **data,
        comments=comments,
        exposure_table=exposure_table_ascii,
        notifications=notifications,
        metrics=NightMetrics(**metrics),
    )


@router.get("/night-logs/{mjd}/email", summary="Email night log")
async def route_get_night_logs_mjd_email(
    mjd: Annotated[
        int,
        Path(description="The MJD for which to retrieve night log."),
    ],
    only_if_not_sent: Annotated[
        bool,
        Query(description="Only send the email if it has not been already been sent."),
    ] = False,
):
    """Emails the night log."""

    mjd = mjd if mjd > 0 else get_sjd("LCO")

    try:
        await email_night_log(mjd, only_if_not_sent=only_if_not_sent)
    except Exception as err:
        if "has already been sent" in str(err):
            return False
        raise

    return True


@router.get("/night-logs/{mjd}/plaintext", summary="Plain-text night log")
async def route_get_night_logs_mjd_plaintext(
    mjd: Annotated[
        int,
        Path(description="The MJD for which to retrieve night log."),
    ],
):
    """Returns the night log as a plain-text string."""

    mjd = mjd if mjd > 0 else get_sjd("LCO")
    data = await get_plaintext_night_log(mjd)

    return data


@router.get("/night-logs/{mjd}/metrics", summary="Night metrics")
async def route_get_night_logs_metrics(
    mjd: Annotated[
        int,
        Path(description="The MJD for which to retrieve night log."),
    ],
):
    """Returns the night metrics."""

    mjd = mjd if mjd > 0 else get_sjd("LCO")
    data = get_night_metrics(mjd)

    return NightMetrics(**data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-11-09
# @Filename: notifications.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import datetime

from typing import Annotated, Any

from fastapi import APIRouter, HTTPException, Path
from pydantic import BaseModel, Field

from sdsstools import get_sjd

from lvmapi.tools.notifications import (
    NotificationLevel,
    create_notification,
    get_notifications,
)


router = APIRouter(prefix="/notifications", tags=["notifications"])


class Notification(BaseModel):
    """An Overwatcher notification."""

    date: Annotated[
        datetime.datetime,
        Field(description="The notification datetime."),
    ]
    message: Annotated[
        str,
        Field(description="The notification message."),
    ]
    level: Annotated[
        str,
        Field(description="The level of the notification."),
    ]
    payload: Annotated[
        dict | None,
        Field(description="The payload of the notification."),
    ] = None


class NotificationPost(BaseModel):
    """Model for the ``/notifications/create`` endpoint payload."""

    message: Annotated[
        str,
        Field(description="The notification message."),
    ]
    level: Annotated[
        str,
        Field(description="The level of the notification."),
    ] = "INFO"
    payload: Annotated[
        dict[str, Any], Field(description="The payload of the notification.")
    ] = {}
    slack_channel: Annotated[
        str | bool | None,
        Field(description="The Slack channel where to send the message."),
    ] = None
    email_on_critical: Annotated[
        bool, Field(description="Whether to send an email if the level is CRITICAL.")
    ] = True
    write_to_database: Annotated[
        bool, Field(description="Whether to write the notification to the database.")
    ] = True
    slack_extra_params: Annotated[
        dict[str, Any],
        Field(description="Extra parameters to pass to the Slack message."),
    ] = {}


@router.get("/{mjd}", summary="Returns notifications for an MJD.")
async def route_get_notifications(
    mjd: Annotated[
        int,
        Path(description="The SJD for which to list notifications. 0 for current SJD."),
    ],
) -> list[Notification]:
    """Returns a list of notifications for an MJD."""

    mjd = mjd if mjd > 0 else get_sjd("LCO")
    notifications = await get_notifications(mjd)

    return [
        Notification(**notification)
        for notification in notifications
        if notification["message"] != "I am alive!"
    ]


@router.post("/create", summary="Create a new notification.")
async def route_post_create_notification(notification: NotificationPost):
    """Creates a new notification, optionally emitting Slack and email messages."""

    try:
        await create_notification(
            notification.message,
            level=NotificationLevel(notification.level.upper()),
            payload=notification.payload,
            slack_channel=notification.slack_channel,
            email_on_critical=notification.email_on_critical,
            write_to_database=notification.write_to_database,
            slack_extra_params=notification.slack_extra_params,
        )
    except Exception as ee:
        raise HTTPException(500, detail=str(ee))



# encoding: utf-8

import os
import pathlib

from sdsstools import get_package_version
from sdsstools.configuration import get_config


# pip package name
NAME = "lvmapi"

# package name should be pip package name
__version__ = get_package_version(path=__file__, package_name=NAME)

internal_config_path = pathlib.Path(__file__).parent / "config.yaml"
config_path = os.getenv("LVMAPI_CONFIG_PATH", None)

config = get_config("lvmapi", config_file=internal_config_path, user_path=config_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-03-24
# @Filename: ephemeris.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import Annotated

from fastapi import APIRouter, Query
from pydantic import BaseModel

from sdsstools import get_sjd

from lvmapi.tools.schedule import get_ephemeris_summary


class EphemerisSummaryOut(BaseModel):
    """Summary of the ephemeris."""

    SJD: int
    request_jd: float
    date: str
    sunset: float
    twilight_end: float
    twilight_start: float
    sunrise: float
    is_night: bool
    is_twilight: bool
    time_to_sunset: float
    time_to_sunrise: float
    moon_illumination: float
    from_file: bool


router = APIRouter(prefix="/ephemeris", tags=["ephemeris"])


@router.get(
    "/summary",
    summary="Summary of the ephemeris",
    response_model=EphemerisSummaryOut,
)
async def route_get_summary(
    sjd: Annotated[
        int | None,
        Query(description="The SJD for which to retrieve the ephemeris summary"),
    ] = None,
):
    """Returns a summary of the ephemeris for an SJD."""

    return get_ephemeris_summary(sjd=sjd)


@router.get("/sjd", summary="Get current SJD")
async def route_get_sjd() -> int:
    return get_sjd("LCO")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-07-25
# @Filename: actors.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import Annotated

from fastapi import APIRouter, Query, Request
from pydantic import BaseModel

from lvmapi import config
from lvmapi.tasks import restart_kubernetes_deployment_task
from lvmapi.tools.logs import get_actor_versions
from lvmapi.tools.rabbitmq import ping_actors


class HealthResponse(BaseModel):
    """Health response for an actor."""

    actor: str
    deployment_name: str
    is_deployed: bool
    ping: bool


router = APIRouter(prefix="/actors", tags=["actors"])


@router.get("/", summary="List of actors")
async def get_actors_route() -> list[str]:
    """Returns a list of actors."""

    return config["actors"]["list"]


@router.get("/health", summary="Actor health")
async def get_actor_health(request: Request) -> list[HealthResponse]:
    """Returns the health of all actors."""

    actors = config["actors"]["list"]
    health: list[HealthResponse] = []

    deployments = request.app.state.kubernetes.list_deployments()
    ping_response = await ping_actors(actors=actors)

    for actor in actors:
        deployment_name = config["actors.actor_to_deployment"][actor]
        ping = ping_response.get(actor, False)
        is_deployed = deployment_name in deployments
        health.append(
            HealthResponse(
                actor=actor,
                deployment_name=deployment_name,
                is_deployed=is_deployed,
                ping=ping,
            )
        )

    return health


@router.get("/ping", summary="Actor ping")
async def get_ping_route(actors: list[str] | None = None) -> dict[str, bool]:
    """Pings a list of actors."""

    return await ping_actors(actors=actors)


@router.get("/restart/{actor}", summary="Restart an actor")
async def get_restart_actor_route(actor: str) -> str:
    """Restarts an actor. Scheduled as a task and returns the task ID"""

    deployment = config["actors.actor_to_deployment"][actor]
    if deployment is None:
        raise ValueError(f"Actor {actor} does not have a deployment.")

    task = await restart_kubernetes_deployment_task.kiq(deployment)
    return task.task_id


@router.get("/stop/{actor}", summary="Stop an actor")
async def get_stop_actor_route(actor: str) -> bool:
    """Stops an actor."""

    from lvmapi.app import app

    deployment = config["actors.actor_to_deployment"][actor]
    if deployment is None:
        raise ValueError(f"Actor {actor} does not have a deployment.")

    app.state.kubernetes.delete_deployment(deployment)

    return True


@router.get("/versions", summary="Get actor versions")
async def get_actor_versions_route(
    actor: Annotated[
        str | None,
        Query(description="Optionally, actor for which to return the version"),
    ] = None,
) -> dict[str, str | None]:
    """Returns the version of an actor."""

    return await get_actor_versions(actor=actor)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-06-14
# @Filename: enclosure.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import enum

from typing import Annotated, Any, Literal

from fastapi import APIRouter, Body, HTTPException, Path
from pydantic import BaseModel, Field, create_model, field_validator

from lvmopstools.devices import read_nps
from sdsstools.utils import GatheringTaskGroup

from lvmapi.auth import AuthDependency
from lvmapi.tasks import move_dome_task
from lvmapi.tools.rabbitmq import send_clu_command


router = APIRouter(prefix="/enclosure", tags=["enclosure"])


class PLCStatus(BaseModel):
    """A model to represent a PLC sensor or function."""

    name: str
    value: int
    labels: list[str] = []

    @field_validator("value", mode="before")
    def cast_value(cls, value: str) -> int:
        return int(value, 16)

    @field_validator("labels", mode="after")
    def trim_labels(cls, value: list[str]) -> list[str]:
        return list(map(str.strip, value))


class O2Status(BaseModel):
    """Status of the O2 sensors."""

    utilities_room: float
    spectrograph_room: float


class EnclosureStatus(BaseModel):
    """A model to represent the status of the enclosure."""

    registers: dict[str, bool | float | int]

    dome_status: PLCStatus
    safety_status: PLCStatus
    lights_status: PLCStatus

    o2_status: O2Status

    cal_lamp_state: dict[str, bool]


class Lights(enum.Enum):
    control_room = "control_room"
    spectrograph_room = "spectrograph_room"
    utilities_room = "utilities_room"
    uma_room = "uma_room"
    telescope_bright = "telescope_bright"
    telescope_red = "telescope_red"
    all = "all"


LightsStatus: type[BaseModel] = create_model(
    "LightsStatus",
    __doc__="Model of lights status",
    **{light: (bool, False) for light in Lights.__members__ if light != "all"},  # type: ignore
)


class NPSBody(BaseModel):
    """Model for NPS data."""

    outlet: Annotated[str, Field(description="The outlet to control or 'all'.")]
    state: Annotated[bool, Field(description="The state to set (on/off).")]


@router.get("")
@router.get("/")
@router.get("/status")
async def status() -> EnclosureStatus:
    """Performs an emergency shutdown of the enclosure and telescopes."""

    try:
        async with GatheringTaskGroup() as group:
            group.create_task(send_clu_command("lvmecp status"))
            group.create_task(route_get_nps(nps="calib"))

        ecp_status, cal_lamp_state = group.results()
    except Exception as ee:
        raise HTTPException(status_code=500, detail=str(ee))

    status_data: dict[str, Any] = {}
    for reply in ecp_status:
        if "registers" in reply:
            status_data["registers"] = reply["registers"]
        elif "dome_status" in reply:
            status_data["dome_status"] = {
                "name": "dome",
                "value": reply["dome_status"],
                "labels": reply["dome_status_labels"].split(","),
            }
        elif "lights" in reply:
            status_data["lights_status"] = {
                "name": "lights",
                "value": reply["lights"],
                "labels": reply["lights_labels"].split(","),
            }
        elif "safety_status" in reply:
            status_data["safety_status"] = {
                "name": "safety",
                "value": reply["safety_status"],
                "labels": reply["safety_status_labels"].split(","),
            }
        elif "o2_percent_utilities" in reply:
            status_data["o2_status"] = {
                "utilities_room": reply["o2_percent_utilities"],
                "spectrograph_room": reply["o2_percent_spectrograph"],
            }

    return EnclosureStatus(**status_data, cal_lamp_state=cal_lamp_state)


@router.get("/open", dependencies=[AuthDependency])
async def open_enclosure():
    """Opens the enclosure. Scheduled as a task."""

    task = await move_dome_task.kiq(direction="open")
    return task.task_id


@router.get("/close", dependencies=[AuthDependency])
async def close_enclosure(force: bool = False):
    """Closes the enclosure. Scheduled as a task."""

    task = await move_dome_task.kiq(direction="close", force=force)
    return task.task_id


@router.get("/stop", dependencies=[AuthDependency])
async def stop_enclosure():
    """Stops the enclosure."""

    try:
        await send_clu_command("lvmecp dome stop")
    except Exception as ee:
        raise HTTPException(status_code=500, detail=str(ee))

    return True


@router.get("/lights", response_model=LightsStatus)
async def get_lights():
    """Returns the status of the lights."""

    try:
        ecp_status = await send_clu_command("lvmecp lights")
    except Exception as ee:
        raise HTTPException(status_code=500, detail=str(ee))

    lights = ecp_status[0]["lights_labels"]

    response_dict: dict[str, bool] = {}
    for light in lights:
        response_dict[light.lower()] = True

    return LightsStatus(**response_dict)


@router.get("/lights/{mode}/{lamp}", dependencies=[AuthDependency])
async def set_lights(
    mode: Literal["on", "off"],
    lamp: Annotated[
        Lights,
        Path(
            description="The lamp to turn on/off. Can also be 'all' (only allowed "
            "with mode='off').",
        ),
    ],
):
    """Turns a lamp on/off."""

    lamp_str = lamp.value

    if mode == "on" or (mode == "off" and lamp_str != "all"):
        if lamp_str == "all":
            raise HTTPException(
                status_code=400,
                detail="Lamp must be specified when turning lights on.",
            )
        await send_clu_command(f"lvmecp lights {mode} {lamp_str}")

    else:
        await asyncio.gather(
            *[
                send_clu_command(f"lvmecp lights off {lamp}")
                for lamp in Lights.__members__
                if lamp != "all"
            ]
        )


@router.get("/nps/{nps}", summary="Reads an NPS")
async def route_get_nps(
    nps: Annotated[str, Path(description="The NPS for which to return the data.")],
) -> dict[str, bool]:
    """Returns the state of a network power switch."""

    try:
        nps_data = await read_nps()
    except Exception as ee:
        raise HTTPException(500, detail=str(ee))

    return {
        outlet.split(".")[1]: nps_data[outlet]["state"]
        for outlet in nps_data
        if outlet.split(".")[0] == nps
    }


@router.put("/nps/{nps}", summary="Sets an NPS switch")
async def route_put_nps(
    nps: Annotated[
        Literal["calib"],
        Path(description="The NPS for which to set the data."),
    ],
    data: Annotated[
        NPSBody,
        Body(description="The NPS data to set."),
    ],
) -> bool:
    """Sets the state of an NPS outlet."""

    command = "on" if data.state else "off"

    try:
        if data.outlet == "all":
            if data.state:
                raise HTTPException(
                    status_code=400,
                    detail="Cannot turn all outlets on.",
                )
            await send_clu_command(f"lvmnps.{nps} all-off")
        else:
            await send_clu_command(f"lvmnps.{nps} {command} {data.outlet}")
    except RuntimeError:
        return False

    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-12
# @Filename: slack.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import Annotated

from fastapi import APIRouter, HTTPException, Query
from pydantic import BaseModel, Field, model_validator

import lvmapi.tools.slack
from lvmapi import config


default_channel = config["slack.default_channel"]
default_user: str = "LVM"


class Message(BaseModel):
    text: Annotated[
        str | None,
        Field(description="Text to be sent"),
    ] = None

    blocks: Annotated[
        list[dict] | None,
        Field(description="Blocks of data to be sent"),
    ] = None

    attachments: Annotated[
        list[dict] | None,
        Field(description="Attachments to be sent"),
    ] = None

    channel: Annotated[
        str | None,
        Field(
            description="Channel where the message will be sent. "
            f"Defaults to {default_channel}."
        ),
    ] = None

    username: Annotated[
        str,
        Field(description="User that sends the message"),
    ] = default_user

    icon_url: Annotated[
        str | None,
        Field(description="URL for the icon to use"),
    ] = None

    mentions: Annotated[
        list[str],
        Field(description="List of user mentions"),
    ] = []

    @model_validator(mode="after")
    def validate_message(self):
        if self.text is None and self.blocks is None:
            raise ValueError("Must specify either text or blocks.")

        if self.text is not None and self.blocks is not None:
            raise ValueError("Cannot specify both text and blocks.")

        return self


router = APIRouter(prefix="/slack", tags=["slack"])


@router.get("/", summary="Slack API")
async def route_get_slack():
    """Not implemented."""

    return {}


@router.post("/message", summary="Send a message to Slack")
async def route_post_message(message: Message) -> None:
    """Sends a message to the Slack channel."""

    try:
        await lvmapi.tools.slack.post_message(
            message.text,
            blocks=message.blocks,
            channel=message.channel,
            username=message.username,
            icon_url=message.icon_url,
            mentions=message.mentions,
            attachments=message.attachments,
        )
    except Exception as err:
        raise HTTPException(500, detail=str(err))

    return None


@router.get("/message", summary="Send a simple text message to Slack")
async def route_get_message(
    text: str = Query(description="Text to be sent"),
    channel: str | None = Query(None, description="Channel where to send the message"),
    username: str = Query(default_user, description="Username to send the message as"),
    icon_url: str | None = Query(None, description="URL for the icon to use"),
    color: str | None = Query(None, description="Color of the message attachment"),
) -> None:
    """Sends a message to the Slack channel."""

    if color is not None:
        attachments = [
            {
                "color": color,
                "text": text,
                "mrkdwn_in": ["text"],
                "fallback": text,
            }
        ]
        text = ""
    else:
        attachments = []

    await route_post_message(
        Message(
            text=text,
            channel=channel,
            username=username,
            icon_url=icon_url,
            attachments=attachments,
        )
    )

    return None


@router.get("/user_id/{user}", summary="Gets the userID for a user name")
async def route_get_user_id(user: str) -> str | None:
    """Gets the ``userID`` of the user whose ``name`` or ``real_name`` matches."""

    try:
        user_id = await lvmapi.tools.slack.get_user_id(user)
    except NameError:
        return None
    except Exception as err:
        raise HTTPException(500, detail=str(err))

    return user_id



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-06-14
# @Filename: gort.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from contextlib import asynccontextmanager

from typing import TYPE_CHECKING, AsyncGenerator

from gort import Gort


if TYPE_CHECKING:
    from fastapi import FastAPI


__all__ = ["get_gort_client"]


@asynccontextmanager
async def get_gort_client(
    app: FastAPI | None = None,
    verbosity: str = "warning",
) -> AsyncGenerator[Gort]:
    """Returns an initialised GORT client.

    Parameters
    ----------
    app
        The FastAPI app. If provided, the GORT client is added to the app state.
        If the state already has a GORT client, checks that it is connected and if
        so returns it.

    Returns
    -------
    gort
        The GORT client.

    """

    gort: Gort

    if app and hasattr(app.state, "gort"):
        gort = app.state.gort
        if not gort.connected:
            await gort.init()
    else:
        gort = await Gort(verbosity=verbosity).init()
        if app:
            app.state.gort = gort

    yield gort



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-07-24
# @Filename: app.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import logging
import os

import taskiq_fastapi
from fastapi import FastAPI, HTTPException, Request

from lvmapi import auth, config
from lvmapi.broker import broker, broker_shutdown, broker_startup
from lvmapi.routers import (
    actors,
    alerts,
    enclosure,
    ephemeris,
    kubernetes,
    logs,
    macros,
    notifications,
    overwatcher,
    slack,
    spectrographs,
    tasks,
    telescopes,
    transparency,
    weather,
)
from lvmapi.tools.kubernetes import Kubernetes


logger = logging.getLogger("uvicorn.error")

if config._CONFIG_FILE is not None:
    logger.info(f"Using configuration from {config._CONFIG_FILE}.")


app = FastAPI(swagger_ui_parameters={"tagsSorter": "alpha"})

app.include_router(auth.router)
app.include_router(telescopes.router)
app.include_router(spectrographs.router)
app.include_router(slack.router)
app.include_router(ephemeris.router)
app.include_router(overwatcher.router)
app.include_router(weather.router)
app.include_router(macros.router)
app.include_router(enclosure.router)
app.include_router(alerts.router)
app.include_router(tasks.router)
app.include_router(kubernetes.router)
app.include_router(actors.router)
app.include_router(logs.router)
app.include_router(notifications.router)
app.include_router(transparency.router)


@app.get("/id")
async def get_id_route(request: Request):
    """Returns the ID of the FastAPI app."""

    return id(request.app)


# Lifecycle events for the broker.
app.add_event_handler("startup", broker_startup)
app.add_event_handler("shutdown", broker_shutdown)

# Integration with FastAPI.
taskiq_fastapi.init(broker, "lvmapi.app:app")

# Add kubernetes API instance to state.
app.state.kubernetes = Kubernetes()

# Fake states for testing.
app.state.use_fake_states = os.environ.get("LVM_USE_FAKE_STATES", "0") != "0"
app.state.fake_states = {
    "wind_alert": False,
    "humidity_alert": False,
    "rain_alert": False,
    "door_alert": False,
    "is_day": False,
}


@app.get("/fake-states/enable", include_in_schema=False)
async def route_get_enable_states():
    """Enable fake states."""

    app.state.use_fake_states = True

    return {"use_fake_states": app.state.use_fake_states}


@app.get("/fake-states/disable", include_in_schema=False)
async def route_get_disable_states():
    """Disable fake states."""

    app.state.use_fake_states = False

    return {"use_fake_states": app.state.use_fake_states}


@app.get("/fake-states/set/{state}/{value}", include_in_schema=False)
async def route_get_set_fake_state(state: str, value: bool):
    """Sets a fake state."""

    if state not in app.state.fake_states:
        raise HTTPException(400, f"Invalid state {state!r}")

    app.state.fake_states[state] = value

    return {"state": state, "value": app.state.fake_states[state]}


@app.get("/fake-states/get", include_in_schema=False)
async def route_get_get_fake_state_all():
    """Queries the value of all fake states."""

    return {
        "use_fake_states": app.state.use_fake_states,
        "fake_States": app.state.fake_states,
    }


@app.get("/fake-states/get/{state}", include_in_schema=False)
async def route_get_get_fake_state(state: str):
    """Queries the value of a fake state."""

    if state not in app.state.fake_states:
        raise HTTPException(400, f"Invalid state {state!r}")

    return {
        "use_fake_states": app.state.use_fake_states,
        "state": state,
        "value": app.state.fake_states[state],
    }


@app.get("/")
def root(request: Request):
    return {}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-11-09
# @Filename: notifications.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import datetime
import enum
import json
import pathlib
import re
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from typing import Any, cast

import psycopg
from jinja2 import Environment, FileSystemLoader
from psycopg.rows import dict_row
from psycopg.sql import SQL, Identifier

from sdsstools import get_sjd

from lvmapi import config
from lvmapi.tools.general import insert_to_database
from lvmapi.tools.slack import post_message


__all__ = ["get_notifications"]


class NotificationLevel(enum.Enum):
    """Allowed notification levels."""

    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


async def fill_notifications_mjd():
    """Fills MJD field for records in the notifications table.

    This is a temporary function to fill the MJD field in the notifications table and
    only needed due to a bug in the Overwatcher which was not filling the MJD field.

    """

    uri = config["database.uri"]
    table = Identifier(*config["database.tables.notification"].split("."))

    # Start by getting a list of records with null MJD.
    query = SQL("SELECT pk, date FROM {table} WHERE mjd IS NULL").format(table=table)
    async with await psycopg.AsyncConnection.connect(uri) as aconn:
        async with aconn.cursor() as acursor:
            await acursor.execute(query)
            records = await acursor.fetchall()

    for pk, date in records:
        sjd = get_sjd("LCO", date=date)
        query = SQL("UPDATE {table} SET mjd = %s WHERE pk = %s").format(table=table)
        async with await psycopg.AsyncConnection.connect(uri) as aconn:
            async with aconn.cursor() as acursor:
                await acursor.execute(query, (sjd, pk))
                await aconn.commit()


async def get_notifications(sjd: int | None = None):
    """Returns the notifications for an SJD."""

    sjd = sjd or get_sjd("LCO")

    uri = config["database.uri"]
    table = Identifier(*config["database.tables.notification"].split("."))

    query = SQL("""
        SELECT * FROM {table}
        WHERE mjd = %s AND message != %s ORDER BY date ASC
    """)

    async with await psycopg.AsyncConnection.connect(uri) as aconn:
        async with aconn.cursor(row_factory=dict_row) as acursor:
            await acursor.execute(query.format(table=table), (sjd, "I am alive!"))
            notifications = await acursor.fetchall()

    return notifications


def format_message_for_db(message: str) -> str:
    """Formats a message for storage in the database.

    Currently the only thing that is changed is replacing the Slack-flavoured
    link style (``<https://example.com|example>``) with a standard Markdown
    link format.

    """

    message = re.sub(
        r"(.*)<(.+)\|([\w\s]+)>",
        r"\1[\3](\2)",
        message,
        flags=re.ASCII | re.IGNORECASE,
    )

    return message


async def create_notification(
    message: str,
    level: NotificationLevel | str = NotificationLevel.INFO,
    payload: dict[str, Any] = {},
    slack_channel: str | bool | None = None,
    email_on_critical: bool = True,
    write_to_database: bool = True,
    slack_extra_params: dict[str, Any] = {},
):
    """Creates a new notification.

    Parameters
    ----------
    message
        The message of the notification. Can be formatted in Markdown.
    level
        The level of the notification.
    payload
        A dictionary with additional information to be stored with the notification.
        This data is not send over email or Slack, only stored in the database.
    slack_channel
        The Slack channel where to send the notification. If not provided, the default
        channel is used. Can be set to false to disable sending the Slack notification.
    email_on_critical
        Whether to send an email if the notification level is ``CRITICAL``.
    write_to_database
        Whether to write the notification to the database.
    slack_extra_params
        A dictionary of extra parameters to pass to ``post_message``.

    """

    date = datetime.datetime.now()
    payload_str = json.dumps(payload)

    if isinstance(level, str):
        level = NotificationLevel(level.upper())
    else:
        level = NotificationLevel(level)

    table = config["database.tables.notification"]

    slack = slack_channel is not False
    email = email_on_critical and level == NotificationLevel.CRITICAL

    if email:
        try:
            await send_critical_error_email(message)
        except Exception as ee:
            print(f"Error sending critical error email: {ee}")

    if slack_channel is not False:
        default_channel: str
        if slack_channel is None or slack_channel is True:
            default_channel = cast(str, config["slack.default_channel"])
        else:
            default_channel = slack_channel

        # We send the message to the default channel plus any other channel that
        # matches the level of the notification.
        channels: set[str] = {default_channel}

        level_channels = cast(dict[str, str], config["slack.level_channels"])
        if level.value in level_channels:
            channels.add(level_channels[level.value])

        # Send Slack message(s)
        for channel in channels:
            mentions = (
                ["@channel"]
                if level == NotificationLevel.CRITICAL
                or level == NotificationLevel.ERROR
                else []
            )
            await post_message(
                message,
                channel=channel,
                mentions=mentions,
                **slack_extra_params,
            )

    if write_to_database:
        message_db = format_message_for_db(message)
        await insert_to_database(
            table,
            [
                {
                    "date": date,
                    "mjd": get_sjd("LCO", date=date),
                    "message": message_db,
                    "level": level.value.lower(),
                    "payload": payload_str,
                    "slack": slack,
                    "email": email,
                }
            ],
            columns=["date", "mjd", "message", "level", "payload", "slack", "email"],
        )


async def send_critical_error_email(message: str):
    """Sends a critical error email."""

    root = pathlib.Path(__file__).parents[1]
    template = root / config["notifications.critical.email_template"]
    loader = FileSystemLoader(template.parent)

    print(template, template.parent)

    env = Environment(
        loader=loader,
        lstrip_blocks=True,
        trim_blocks=True,
    )
    html_template = env.get_template(template.name)

    html_message = html_template.render(message=message.strip())

    recipients = config["notifications.critical.email_recipients"]
    from_address = config["notifications.critical.email_from"]

    email_server = config["notifications.critical.email_server"]
    email_host, *email_rest = email_server.split(":")
    email_port: int = 0
    if len(email_rest) == 1:
        email_port = int(email_rest[0])

    email_reply_to = config["notifications.critical.email_reply_to"]

    msg = MIMEMultipart("alternative" if html_message else "mixed")
    msg["Subject"] = "LVM Critical Alert"
    msg["From"] = from_address
    msg["To"] = ", ".join(recipients)
    msg["Reply-To"] = email_reply_to

    plaintext_email = f"""A critical alert was raised in the LVM system.

The error message is shown below:

{message}

    """
    msg.attach(MIMEText(plaintext_email, "plain"))

    html = MIMEText(html_message, "html")
    msg.attach(html)

    with smtplib.SMTP(host=email_host, port=email_port) as smtp:
        smtp.sendmail(from_address, recipients, msg.as_string())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-12
# @Filename: rabbitmq.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

from lvmopstools.clu import CluClient, send_clu_command
from sdsstools.utils import GatheringTaskGroup

from lvmapi import config


if TYPE_CHECKING:
    pass


__all__ = ["CluClient", "send_clu_command", "ping_actors"]


async def ping_actors(actors: list[str] | None = None):
    """Pings all actors and returns a list of replies."""

    actors = actors or config["actors"]["list"]
    assert actors is not None

    async with CluClient() as client:
        async with GatheringTaskGroup() as group:
            for actor in actors:
                group.create_task(client.send_command(actor, "ping"))

    results = group.results()

    actor_to_pong: dict[str, bool] = {}
    for idx, result in enumerate(results):
        replies = result.replies

        if "text" in replies[-1].message and replies[-1].message["text"] == "Pong.":
            actor_to_pong[actors[idx]] = True
        else:
            actor_to_pong[actors[idx]] = False

    return actor_to_pong



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-11-11
# @Filename: transparency.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import polars

from lvmapi.tools.influxdb import query_influxdb


async def get_transparency(start_time: float, end_time: float):
    """Returns transparency measurements."""

    query = rf"""
from(bucket: "actors")
  |> range(start: {int(start_time)}, stop: {int(end_time)})
  |> filter(fn: (r) => (r["_measurement"] =~ /lvm\.[a-z]+\.guider/) and
                        (r["_field"] == "measured_pointing.zero_point"))
  |> yield(name: "mean")
"""

    data = await query_influxdb(query)

    # Clean up the dataframe.
    data = data.select(
        time=polars.col._time,
        telescope=polars.col._measurement.str.extract(r"lvm\.([a-z]+)\.guider"),
        zero_point=polars.col._value,
    )

    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-03-25
# @Filename: redis.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import redis


__all__ = ["get_redis_connection"]


def get_redis_connection() -> redis.Redis:
    """Returns a connection to the Redis server."""

    return redis.Redis(host="10.8.38.26", port=6379, db=0, decode_responses=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-11
# @Filename: cryostats.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import datetime
import pathlib
import re
import warnings

from typing import Annotated, get_args

import polars
from fastapi import APIRouter, HTTPException, Path, Query
from pydantic import BaseModel, Field

from lvmopstools.devices.ion import IonPumpDict
from lvmopstools.devices.nps import read_nps

from lvmapi.tools.spectrograph import (
    FillMetadataReturn,
    exposure_etr,
    get_fill_list,
    read_ion_pumps,
    read_thermistors,
    read_thermistors_influxdb,
    register_ln2_fill,
    retrieve_fill_measurements,
    retrieve_fill_metadata,
    spectrograph_mechanics,
    spectrograph_pressures,
    spectrograph_status,
    spectrograph_temperatures,
    spectrograph_temperatures_history,
)
from lvmapi.types import CamSpec, Sensors, SpecStatus, Spectrographs


class SpecStatusResponse(BaseModel):
    status: dict[Spectrographs, SpecStatus] = Field(
        ...,
        description="The status of each spectrograph",
    )
    last_exposure_no: int = Field(
        ...,
        description="The last exposure number",
    )
    total_exposure_time: float | None = Field(
        default=None,
        description="The total exposure time, including readout",
    )
    exposure_etr: float | None = Field(
        default=None,
        description="The estimated time remaining for the current exposure, "
        "including readout",
    )


class IonPumpResponse(BaseModel):
    """The status of the ion pumps."""

    pressure: float | None = Field(
        ...,
        description="The pressure reported by the ion pump in Torr",
    )
    on: bool | None = Field(
        ...,
        description="Whether the ion pump is on",
    )


class CryostatsResponse(BaseModel):
    """Reponse model for ``/cryostats``."""

    ln2_temperatures: Annotated[
        dict[str, float | None],
        Field(description="LN2 temperatures"),
    ]
    pressures: Annotated[
        dict[str, float | None],
        Field(
            description="Pressures in Torr as reported by the "
            "cryostat pressure transceivers"
        ),
    ]
    pressures_ion: Annotated[
        dict[str, IonPumpDict],
        Field(description="Pressures in Torr as reported by the ion pumps"),
    ]
    thermistors: Annotated[
        dict[str, bool | None],
        Field(description="Whether the thermistors are sensing LN2"),
    ]
    purging: Annotated[
        bool | None,
        Field(description="Is the purge valve open?"),
    ] = False
    filling: Annotated[
        dict[str, bool | None],
        Field(description="Whether the cryostat are filling"),
    ]


class FillDataModel(BaseModel):
    """Response for the ``/fills/data`` endpoint."""

    time: Annotated[datetime.datetime, Field(description="Time of the measurement")]

    # There are many more fields so we just allow them all.
    model_config = {"extra": "allow"}


class RegisterFillPostModel(BaseModel):
    """Request model for the ``/fills/register`` endpoint."""

    action: Annotated[
        str,
        Field(description="LN2 action performed"),
    ]
    complete: Annotated[
        bool,
        Field(description="Is the action complete?"),
    ]
    start_time: Annotated[
        str | None,
        Field(description="Start time of the action"),
    ]
    end_time: Annotated[
        str | None,
        Field(description="End time of the action"),
    ]
    purge_start: Annotated[
        str | None,
        Field(description="Purge start time"),
    ]
    purge_complete: Annotated[
        str | None,
        Field(description="Purge finish time"),
    ]
    fill_start: Annotated[
        str | None,
        Field(description="Fill start time"),
    ]
    fill_complete: Annotated[
        str | None,
        Field(description="Fill finish time"),
    ]
    fail_time: Annotated[
        str | None,
        Field(description="Time of failure"),
    ]
    abort_time: Annotated[
        str | None,
        Field(description="Time of abort"),
    ]
    failed: Annotated[
        bool,
        Field(description="Did the action fail?"),
    ]
    aborted: Annotated[
        bool,
        Field(description="Was the action aborted?"),
    ]
    error: Annotated[
        str | None,
        Field(description="Error message"),
    ]
    log_file: Annotated[
        str | None,
        Field(description="Path to the log file"),
    ]
    json_file: Annotated[
        str | None,
        Field(description="Path to the JSON file"),
    ]
    configuration: Annotated[
        dict | None,
        Field(description="Configuration data"),
    ]
    log_data: Annotated[
        list[dict] | None,
        Field(description="Log data in JSON format"),
    ]
    plot_paths: Annotated[
        dict[str, str] | None,
        Field(description="Paths to plots"),
    ]
    valve_times: Annotated[
        dict[str, dict[str, str | bool | None]] | None,
        Field(description="Valve open/close times"),
    ]
    pk: Annotated[
        int | None,
        Field(description="Primary key of the record to update."),
    ]


router = APIRouter(prefix="/spectrographs", tags=["spectrographs"])


@router.get("/", summary="List of spectrographs")
async def route_get_sectrographs() -> list[str]:
    """Returns the list of cryostats."""

    return list(get_args(Spectrographs))


@router.get(
    "/status",
    summary="Returns the spectrograph status",
    response_model=SpecStatusResponse,
)
async def route_get_status() -> SpecStatusResponse:
    """Returns the spectrograph status."""

    spec_status = await spectrograph_status()

    return SpecStatusResponse(
        status=spec_status["status"],
        last_exposure_no=spec_status["last_exposure_no"],
        exposure_etr=spec_status["etr"][0],
        total_exposure_time=spec_status["etr"][1],
    )


@router.get("/etr", summary="Estimated time remaining")
async def route_get_etr():
    """The estimated time remaining for the current exposure, including readout."""

    return await exposure_etr()


@router.get("/temperatures", summary="Cryostat temperatures")
async def route_get_temperatures(
    start: str = Query("-30m", description="Flux-compatible start time"),
    stop: str = Query("now()", description="Flux-compatible stop time"),
    camera: CamSpec | None = Query(None, description="Camera to return, or all"),
    sensor: Sensors | None = Query(None, description="Sensor to return, or all"),
    last: bool = Query(False, description="Return only the last set of temperatures"),
) -> list[dict]:
    """Returns the temperatures of one or multiple cryostats."""

    df = await spectrograph_temperatures_history(
        start,
        stop,
        camera=camera,
        sensor=sensor,
    )

    if last:
        df = df.group_by(["camera", "sensor"]).last()

    df = df.with_columns(time=polars.col.time.dt.to_string("%Y-%m-%dT%H:%M:%S.%3f"))
    df = df.select(polars.col(["time", "camera", "sensor", "temperature"]))
    df = df.sort(["camera", "sensor"])

    return df.to_dicts()


@router.get("/thermistors")
@router.get(
    "/thermistors/{thermistor}",
    summary="Reads the thermistors",
    response_model=list[dict] | dict[str, bool] | bool,
)
async def route_get_thermistors(
    thermistor: str | None = None,
    interval: float | None = Query(None, description="Interval in seconds"),
):
    """Reads the thermistors and returns their states."""

    data = await read_thermistors_influxdb(interval)

    if isinstance(data, polars.DataFrame):
        data = data.with_columns(
            time=polars.col.time.dt.to_string("%Y-%m-%dT%H:%M:%S.%3f")
        )

        if thermistor is not None:
            data = data.filter(polars.col.channel == thermistor)

        return data.to_dicts()

    if thermistor:
        return data[thermistor]

    return data


@router.get("/ion")
async def route_get_ion() -> dict[str, IonPumpResponse]:
    """Reads the current values of the ion pumps."""

    ion_pump_data = await read_ion_pumps()

    return {key: IonPumpResponse(**value) for key, value in ion_pump_data.items()}


@router.get("/cryostats", summary="Cryostat information")
async def route_get_cryostats():
    """Returns cryostat information regarding LN2 fills and pressures."""

    temperatures_response = await spectrograph_temperatures(ignore_errors=True)
    ln2_temperaures = {
        camera.split("_")[0]: temp
        for camera, temp in temperatures_response.items()
        if camera.endswith("ln2")
    }

    pressures = await spectrograph_pressures(ignore_errors=True)
    pressures_ion = await read_ion_pumps()
    thermistors = await read_thermistors()

    purging = False
    filling = {}

    try:
        nps = await read_nps()
        purging = nps["sp1.purge"]["state"]
        for key, value in nps.items():
            if match := re.match(r"sp[1-3]\.([rzb][1-3])", key):
                filling[match.group(1)] = value["state"]
    except Exception:
        warnings.warn("Error retrieving NPS information.")

    return CryostatsResponse(
        ln2_temperatures=ln2_temperaures,
        pressures=pressures,
        pressures_ion=pressures_ion,
        thermistors=thermistors,
        purging=purging,
        filling=filling,
    )


@router.get("/fills/running", summary="Is a fill currently running?")
async def route_get_fills_running() -> bool:
    """Returns whether an LN2 fill is currently running."""

    return pathlib.Path("/data/lvmcryo.lock").exists()


@router.get("/fills/measurements", summary="Cryostat fill measurements")
async def route_get_fills_measurements(
    start_time: int,
    end_time: int,
) -> list[FillDataModel]:
    """Returns cryostat fill measurements."""

    if start_time < 0 or end_time < 0 or start_time > end_time:
        raise HTTPException(400, detail="Invalid start or end times.")

    data = await retrieve_fill_measurements(start_time, end_time)
    data_dict = data.to_dicts()

    return [FillDataModel(**row) for row in data_dict]


@router.post("/fills/register", summary="Register an LN2 fill")
async def route_post_fills_register(data: RegisterFillPostModel) -> int:
    """Registers or updates an LN2 fill in the database."""

    try:
        pk = await register_ln2_fill(**data.model_dump())
    except Exception as ee:
        raise HTTPException(500, detail=str(ee))

    return pk


@router.get("/fills/list", summary="List LN2 fills and record PKs")
async def route_get_fills_list() -> dict[int, str]:
    """Returns a mapping of LN2 fill PK to start time."""

    return await get_fill_list()


@router.get("/fills/{pk}/metadata", summary="Get LN2 fill metadata")
async def route_get_fills_metadata(
    pk: Annotated[int, Path(description="Primary key of the LN2 fill record")],
    transparent_plots: Annotated[
        bool,
        Query(description="Return transparent metadata?"),
    ] = False,
    as_base64: Annotated[
        bool,
        Query(description="Return plots as base64-encoded strings?"),
    ] = False,
) -> FillMetadataReturn:
    """Returns the metadata for an LN2 fill."""

    try:
        return await retrieve_fill_metadata(
            pk,
            transparent_plots=transparent_plots,
            as_base64=as_base64,
        )
    except ValueError as err:
        if "Cannnot find LN2 fill with pk" in str(err):
            raise HTTPException(400, detail="Invalid primary key.")
        raise


@router.get("/{spectrograph}", summary="Cryostat basic information")
@router.get("/{spectrograph}/summary", summary="Cryostat basic information")
async def route_get_summary(
    spectrograph: Spectrographs,
    mechs: bool = Query(False, description="Return mechanics information?"),
) -> dict[str, float | str | None]:
    """Retrieves current spectrograph information (temperature, pressure, etc.)"""

    try:
        temps_response = await spectrograph_temperatures(
            spectrograph,
            ignore_errors=True,
        )
        pres_reponse = await spectrograph_pressures(
            spectrograph,
            ignore_errors=True,
        )

        if mechs:
            mechs_response = await spectrograph_mechanics(
                spectrograph,
                ignore_errors=True,
            )
        else:
            mechs_response = {}

    except Exception:
        raise HTTPException(500, detail="Error retrieving cryostat information.")

    return temps_response | pres_reponse | mechs_response



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-03-26
# @Filename: weather.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import time

import httpx
import polars
from astropy.time import Time, TimeDelta


__all__ = ["get_weather_data"]


WEATHER_URL = "http://dataservice.lco.cl/vaisala/data"


async def get_weather_data(
    end_time: str | None = None,
    start_time: str | float = 3600,
    station="DuPont",
):
    """Returns weather data from the Vaisala weather station.

    Parameters
    ----------
    end_time
        The end time for the query. If not provided, the current time is used.
        The time can be an Astropy Time object or a string in ISO format.
    start_time
        The start time for the query. The time can be an Astropy Time object,
        a string in ISO format, or a float indicating with a time delta in seconds
        with respect to ``end_time``.

    Returns
    -------
    dataframe
        A Polars dataframe with the time series weather data.

    """

    if station not in ["DuPont", "C40", "Magellan"]:
        raise ValueError("station must be one of 'DuPont', 'C40', or 'Magellan'.")

    if end_time is None:
        end_time_ap = Time.now()
    elif isinstance(end_time, str):
        end_time_ap = Time(end_time)
    elif isinstance(end_time, Time):
        end_time_ap = end_time
    else:
        raise ValueError("end_time must be a string or an Astropy Time object.")

    if isinstance(start_time, str):
        start_time_ap = Time(start_time)
    elif isinstance(start_time, (int, float)):
        start_time_ap = end_time_ap - TimeDelta(start_time, format="sec")
    elif isinstance(start_time, Time):
        start_time_ap = start_time
    else:
        raise ValueError(
            "start_time must be a string, a time delta in seconds, "
            "or an Astropy Time object."
        )

    end_time_ap.precision = 0
    start_time_ap.precision = 0

    async with httpx.AsyncClient() as client:
        response = await client.get(
            WEATHER_URL,
            params={
                "start_ts": str(start_time_ap.iso),
                "end_ts": str(end_time_ap.iso),
                "station": station,
            },
        )

        if response.status_code != 200:
            raise ValueError(f"Failed to get weather data: {response.text}")

        data = response.json()

        if "Error" in data:
            raise ValueError(f"Failed to get weather data: {data['Error']}")
        elif "results" not in data or data["results"] is None:
            raise ValueError("Failed to get weather data: no results found.")

    results = data["results"]

    df = polars.DataFrame(results)
    df = df.with_columns(
        ts=polars.col("ts").str.to_datetime(time_unit="ms"),
        station=polars.lit(station, polars.String),
    )

    # Delete rows with all null values.
    df = df.filter(~polars.all_horizontal(polars.exclude("ts", "station").is_null()))

    # Sort by timestamp
    df = df.sort("ts")

    # Convert wind speeds to mph (the LCO API returns km/h)
    df = df.with_columns(polars.selectors.starts_with("wind_") / 1.60934)

    # Calculate rolling means for average wind speed and gusts every 5m, 10m, 30m
    window_sizes = ["5m", "10m", "30m"]
    df = df.with_columns(
        **{
            f"wind_speed_avg_{ws}": polars.col.wind_speed_avg.rolling_mean_by(
                by="ts",
                window_size=ws,
            )
            for ws in window_sizes
        },
        **{
            f"wind_gust_{ws}": polars.col.wind_speed_max.rolling_max_by(
                by="ts",
                window_size=ws,
            )
            for ws in window_sizes
        },
        **{
            f"wind_dir_avg_{ws}": polars.col.wind_dir_avg.rolling_mean_by(
                by="ts",
                window_size=ws,
            )
            for ws in window_sizes
        },
    )

    # Add simple dew point.
    df = df.with_columns(
        dew_point=polars.col.temperature
        - ((100 - polars.col.relative_humidity) / 5.0).round(2)
    )

    return df


def is_measurament_safe(
    data: polars.DataFrame,
    measurement: str,
    threshold: float,
    window: int = 30,
    rolling_average_window: int = 30,
    reopen_value: float | None = None,
):
    """Determines whether an alert should be raised for a given measurement.

    An alert will be issued if the rolling average value of the ``measurement``
    (a column in ``data``) over the last ``window`` seconds is above the
    ``threshold``. Once the alert has been raised  the value of the ``measurement``
    must fall below the ``reopen_value`` to close the alert (defaults to the same
    ``threshold`` value) in a rolling.

    ``window`` and ``rolling_average_window`` are in minutes.

    Returns
    -------
    result
        A boolean indicating whether the measurement is safe. `True` means the
        measurement is in a valid, safe range.

    """

    if measurement not in data.columns:
        raise ValueError(f"Measurement {measurement} not found in data.")

    reopen_value = reopen_value or threshold

    data = data.select(polars.col.ts, polars.col(measurement))
    data = data.with_columns(
        average=polars.col(measurement).rolling_mean_by(
            by="ts",
            window_size=f"{rolling_average_window}m",
        )
    )

    # Get data from the last window`.
    now = time.time()
    data_window = data.filter(polars.col.ts.dt.timestamp("ms") > (now - window * 60))

    # If any of the values in the last "window" is above the threshold, it's unsafe.
    if (data_window["average"] >= threshold).any():
        return False

    # If all the values in the last "window" are below the reopen threshold, it's safe.
    if (data_window["average"] < reopen_value).all():
        return True

    # The last case is if the values in the last "window" are between the reopen and
    # the threshold values. We want to avoid the returned value flipping from true
    # to false in a quick manner. We check the previous "window" minutes to see if
    # the alert was raised at any point. If so, we require the current window to
    # be below the reopen value. Otherwise, we consider it's safe.

    prev_window = data.filter(
        polars.col.ts.dt.timestamp("ms") > (now - 2 * window * 60),
        polars.col.ts.dt.timestamp("ms") < (now - window * 60),
    )
    if (prev_window["average"] >= threshold).any():
        return (data_window["average"] < reopen_value).all()

    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-19
# @Filename: auth.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

# Mostly adapted from https://fastapi.tiangolo.com/tutorial/security/oauth2-jwt/

from __future__ import annotations

import ipaddress
import os
from datetime import UTC, datetime, timedelta

from typing import Annotated

from fastapi import APIRouter, Depends, HTTPException, Request, status
from fastapi.security import OAuth2PasswordBearer, OAuth2PasswordRequestForm
from jose import JWTError, jwt
from passlib.context import CryptContext
from pydantic import BaseModel


__all__ = ["validate_token", "Token", "AuthDependency"]


# Adapted from https://fastapi.tiangolo.com/tutorial/security/oauth2-jwt/

# To get a string like this run: openssl rand -hex 32
# Secret key and password are expected to be in environment variables.
# For production deployment in Kubernetes with use an opaque secret.
SECRET_KEY = os.getenv("SECRET_KEY")
ALGORITHM = "HS256"
ACCESS_TOKEN_EXPIRE_MINUTES = 60 * 24 * 180  # 180 days

CREDENTIALS_EXCEPTION = HTTPException(
    status_code=status.HTTP_401_UNAUTHORIZED,
    detail="Could not validate credentials",
    headers={"WWW-Authenticate": "Bearer"},
)

# The subnet_or_token allows requests from this domain to bypass the token validation.
ALLOW_SUBNET: str = "10.8.38.0/24"


class Token(BaseModel):
    access_token: str
    token_type: str


pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")
oauth2_scheme = OAuth2PasswordBearer(tokenUrl="/auth/login")

router = APIRouter(prefix="/auth", tags=["auth"])


def verify_password(plain_password: str):
    """Verifies that the password passed matches the hashed one stored in the envvar."""

    PASSWORD = os.getenv("LVMAPI_PASSWORD")  # Already hashed.
    if not PASSWORD:
        return False

    return pwd_context.verify(plain_password, PASSWORD)


def get_password_hash(password: str):
    """Creates a hashed password."""

    return pwd_context.hash(password)


def authenticate(password: str):
    """Authenticates the user."""

    if not verify_password(password):
        return False

    return True


def create_access_token(data: dict, expires_delta: timedelta | None = None):
    """Creates an access token with expiration date."""

    if SECRET_KEY is None:
        raise CREDENTIALS_EXCEPTION

    to_encode = data.copy()
    if expires_delta:
        expire = datetime.now(UTC) + expires_delta
    else:
        expire = datetime.now(UTC) + timedelta(minutes=15)

    to_encode.update({"exp": expire})
    encoded_jwt = jwt.encode(to_encode, SECRET_KEY, algorithm=ALGORITHM)

    return encoded_jwt


async def validate_token(token: Annotated[str, Depends(oauth2_scheme)]):
    """Validates a token."""

    if SECRET_KEY is None:
        raise CREDENTIALS_EXCEPTION

    try:
        payload = jwt.decode(token, SECRET_KEY, algorithms=[ALGORITHM])
        authorised: str | None = payload.get("authorised")
        if authorised is None:
            raise CREDENTIALS_EXCEPTION
    except JWTError:
        raise CREDENTIALS_EXCEPTION

    return {"authorised": True}


async def subnet_or_token(request: Request):
    """Validates a token or allows localhost."""

    if request.client:
        host = request.client.host
        if ipaddress.ip_address(host) in ipaddress.ip_network(ALLOW_SUBNET):
            return {"authorised": True}

    token = await oauth2_scheme(request)
    if not token:
        raise CREDENTIALS_EXCEPTION

    return await validate_token(token)


@router.post("/login", response_model=Token)
async def login_for_access_token(
    form_data: Annotated[OAuth2PasswordRequestForm, Depends()],
):
    """Validates the password and generates a temporary token.

    The OAuth2 standard requires an ``username`` and ``password`` but
    here only the password is used. The username string is required but
    ignored.

    """

    if not authenticate(form_data.password):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect password.",
            headers={"WWW-Authenticate": "Bearer"},
        )

    access_token_expires = timedelta(minutes=ACCESS_TOKEN_EXPIRE_MINUTES)
    access_token = create_access_token(
        data={"authorised": True},
        expires_delta=access_token_expires,
    )

    return {"access_token": access_token, "token_type": "bearer"}


@router.get("/test", dependencies=[Depends(validate_token)])
async def route_get_test():
    """A simple route to test validation."""

    return True


AuthDependency = Depends(validate_token)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-07-25
# @Filename: broker.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import os

import aio_pika
from taskiq_aio_pika import AioPikaBroker
from taskiq_redis import RedisAsyncResultBackend


__all__ = ["broker"]


# The RabbitMQ queue name for the broker and workers to use.
# This allows to create different pools of workers for dev and production.
QUEUE_NAME: str = os.getenv("TASKIQ_QUEUE_NAME", "lvmapi")

EXCHANE_NAME = "taskiq-dev" if "dev" in QUEUE_NAME else "taskiq"


async def broker_startup():
    """Start broker on startup."""

    # Purge all messages from the queue before starting.
    connection = await aio_pika.connect_robust("amqp://guest:guest@127.0.0.1/")

    async with connection:
        channel = await connection.channel()
        await channel.set_qos(prefetch_count=10)

        try:
            queue = await channel.declare_queue(
                QUEUE_NAME,
                arguments={
                    "x-dead-letter-exchange": "",
                    "x-dead-letter-routing-key": "lvmapi-dev.dead_letter",
                },
            )
            await queue.purge()
        except Exception:
            pass

    if not broker.is_worker_process:
        await broker.startup()


async def broker_shutdown():
    """Shut down broker."""

    if not broker.is_worker_process:
        await broker.shutdown()


# TaskIQ broker.
backend = RedisAsyncResultBackend("redis://localhost")
broker = AioPikaBroker(
    queue_name=QUEUE_NAME,
    exchange_name=EXCHANE_NAME,
).with_result_backend(backend)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-08-29
# @Filename: telescopes.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import Any

from lvmapi.tools.rabbitmq import CluClient
from lvmapi.types import Telescopes


async def get_telescope_status(telescope: Telescopes) -> dict[str, Any]:
    """Gets the status of a telescope."""

    async with CluClient() as client:
        status_cmd = await client.send_command(f"lvm.{telescope}.pwi", "status")

    if status_cmd.status.did_fail:
        raise ValueError(f"Failed to get status for telescope {telescope!r}.")

    return {
        "ra": status_cmd.replies.get("ra_j2000_hours") * 15.0,
        "dec": status_cmd.replies.get("dec_apparent_degs"),
        "alt": status_cmd.replies.get("altitude_degs"),
        "az": status_cmd.replies.get("azimuth_degs"),
        "is_tracking": status_cmd.replies.get("is_tracking"),
        "is_connected": status_cmd.replies.get("is_connected"),
        "is_slewing": status_cmd.replies.get("is_slewing"),
        "is_enabled": status_cmd.replies.get("is_enabled"),
    }


def is_telescope_parked(data: dict[str, Any]):
    """Returns whether a telescope is parked."""

    park_position = (-60, 90)
    if (
        not data["reachable"]
        or not data["is_connected"]
        or not data["alt"]
        or not data["az"]
    ):
        return None

    if data["is_slewing"] or data["is_tracking"]:
        return False

    alt_diff = abs(data["alt"] - park_position[0])
    az_diff = abs(data["az"] - park_position[1])

    if alt_diff > 5 or az_diff > 5:
        return False

    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-19
# @Filename: influxdb.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import json
import os
import warnings

import polars

from lvmapi import config


__all__ = ["query_influxdb"]


async def query_influxdb(query: str) -> polars.DataFrame:
    """Runs a query in InfluxDB and returns a Polars dataframe."""

    from influxdb_client.client.influxdb_client_async import InfluxDBClientAsync
    from influxdb_client.client.warnings import MissingPivotFunction

    warnings.simplefilter("ignore", MissingPivotFunction)  # noqa: F821

    host = config["influxdb.host"]
    port = config["influxdb.port"]

    token = config["influxdb.token"] or os.environ.get("INFLUXDB_V2_TOKEN")
    if token is None:
        raise ValueError("$INFLUXDB_V2_TOKEN not defined.")

    async with InfluxDBClientAsync(
        url=f"http://{host}:{port}",
        token=token,
        org=config["influxdb.org"],
    ) as client:
        if not (await client.ping()):
            raise RuntimeError("InfluxDB client failed to connect.")

        api = client.query_api()

        query_results = await api.query(query)

    df = polars.DataFrame(json.loads(query_results.to_json()))

    if len(df) > 0:
        df = df.with_columns(polars.col._time.cast(polars.Datetime("ms")))

    return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-07-25
# @Filename: kubernetes.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from fastapi import APIRouter, Request
from pydantic import BaseModel

from lvmapi.tasks import restart_kubernetes_deployment_task


if TYPE_CHECKING:
    from lvmapi.tools.kubernetes import Kubernetes


class DeploymentInfoResponse(BaseModel):
    """Response for deployment info."""

    api_version: str | None
    kind: str | None
    metadata: dict[str, Any]
    status: dict[str, Any]


router = APIRouter(prefix="/kubernetes", tags=["kubernetes"])


@router.get("/deployments", summary="Lists the deployments")
@router.get("/deployments/list", summary="Lists the deployments")
async def route_get_list_deployments(request: Request) -> list[str]:
    """Lists the deployments in all namespaces."""

    kube: Kubernetes = request.app.state.kubernetes

    return kube.list_deployments()


@router.get("/deployments/{deployment}/restart", summary="Restart deployment")
async def route_get_restart_deployment(deployment: str) -> str:
    """Restarts a deployment. Scheduled as a task (returns task ID)."""

    task = await restart_kubernetes_deployment_task.kiq(deployment)
    return task.task_id


@router.get("/deployments/{deployment}/info", summary="Get deployment info")
async def route_get_deployment_info(request: Request, deployment: str):
    """Returns information about a deployment."""

    kube: Kubernetes = request.app.state.kubernetes

    info = kube.get_deployment_info(deployment)

    return DeploymentInfoResponse(**info)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-12
# @Filename: slack.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import os
import re

from typing import Sequence

from aiocache import cached
from slack_sdk.errors import SlackApiError
from slack_sdk.web.async_client import AsyncWebClient

from lvmapi import config


__all__ = ["post_message", "get_user_id"]


def get_api_client(token: str | None = None):
    """Gets a Slack API client."""

    token = token or config["slack.token"] or os.environ["SLACK_API_TOKEN"]

    return AsyncWebClient(token=token)


async def format_mentions(text: str | None, mentions: list[str]) -> str | None:
    """Formats a text message with mentions."""

    if not text:
        return text

    if len(mentions) > 0:
        for mention in mentions[::-1]:
            if mention[0] != "@":
                mention = f"@{mention}"
            if mention not in text:
                text = f"{mention} {text}"

    # Replace @channel, @here, ... with the API format <!here>.
    text = re.sub(r"(\s|^)@(here|channel|everone)(\s|$)", r"\1<!here>\3", text)

    # The remaining mentions should be users. But in the API these need to be
    # <@XXXX> where XXXX is the user ID and not the username.
    users: list[str] = re.findall(r"(?:\s|^)@([a-zA-Z_]+)(?:\s|$)", text)

    for user in users:
        try:
            user_id = await get_user_id(user)
        except NameError:
            continue
        text = text.replace(f"@{user}", f"<@{user_id}>")

    return text


async def post_message(
    text: str | None = None,
    blocks: Sequence[dict] | None = None,
    channel: str | None = None,
    mentions: list[str] = [],
    **kwargs,
):
    """Posts a message to Slack.

    Parameters
    ----------
    text
        Plain text to send to the Slack channel.
    blocks
        A list of blocks to send to the Slack channel. These follow the Slack
        API format for blocks. Incompatible with ``text``.
    channel
        The channel in the SSDS-V workspace where to send the message.
    mentions
        A list of users to mention in the message.

    """

    if text is None and blocks is None:
        raise ValueError("Must specify either text or blocks.")

    if text is not None and blocks is not None:
        raise ValueError("Cannot specify both text and blocks.")

    channel = channel or config["slack.default_channel"]
    assert channel is not None

    client = get_api_client()

    try:
        text = await format_mentions(text, mentions)
        await client.chat_postMessage(
            channel=channel,
            text=text,
            blocks=blocks,
            **kwargs,
        )
    except SlackApiError as e:
        raise RuntimeError(f"Slack returned an error: {e.response['error']}")


@cached(ttl=120)
async def get_user_list():
    """Returns the list of users in the workspace.

    This function is cached because Slack limits the requests for this route.

    """

    client = get_api_client()

    try:
        users_list = await client.users_list()
        if users_list["ok"] is False:
            err = "users_list returned ok=false"
            raise SlackApiError(err, response={"error": err})

        return users_list

    except SlackApiError as e:
        raise RuntimeError(f"Slack returned an error: {e.response['error']}")


async def get_user_id(name: str):
    """Gets the ``userID`` of the user display name matches ``name``."""

    users_list = await get_user_list()

    for member in users_list["members"]:
        if "profile" not in member or "display_name" not in member["profile"]:
            continue

        if (
            member["profile"]["display_name"] == name
            or member["profile"]["display_name_normalized"] == name
        ):
            return member["id"]

    raise NameError(f"User {name} not found.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-07-12
# @Filename: alerts.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from lvmapi.tools.rabbitmq import CluClient
from lvmapi.tools.spectrograph import spectrograph_temperatures_history


__all__ = ["spec_temperature_alerts", "enclosure_alerts"]


async def spec_temperature_alerts(
    start: str = "-5m",
    ccd_threshold: float = -85,
    ln2_threshold: float = -160,
):
    """Returns a dictionary of spectrograph temperature alerts for each camera.

    Parameters
    ----------
    start
        How far back to query for spectrograph data. Must use the InfluxDB start time
        query format. The stop time for the query is always ``now()``.
    ccd_threshold
        The threshold for the CCD temperature. If the average temperature of the data
        returned for a camera is above this value, the camera is set as alerted.
    ln2_threshold
        As ``ccd_threshold`` for the LN2 can temperature.

    Returns
    -------
    temp_alerts_dict
        A dictionary of alerts. For each camera, the dictionary contains a boolean
        key for the CCD and LN2 temperature alerts, for example
        ``{'r1_ccd': False, 'r1_ln2': False, 'b1_ccd': True, ...}``.

    """

    temperatures = await spectrograph_temperatures_history(start)

    temp_alerts_dict: dict[str, bool] = {}

    for (camera, sensor), gg in temperatures.group_by(["camera", "sensor"]):
        if len(gg) == 0:
            temp_alerts_dict[f"{camera}_{sensor}"] = False
            continue

        mean_temp = float(gg["temperature"].to_numpy().mean())

        if sensor == "ccd":
            temp_alerts_dict[f"{camera}_ccd"] = mean_temp > ccd_threshold
        elif sensor == "ln2":
            temp_alerts_dict[f"{camera}_ln2"] = mean_temp > ln2_threshold

    temp_alerts_dict = {key: temp_alerts_dict[key] for key in sorted(temp_alerts_dict)}

    return temp_alerts_dict


async def enclosure_alerts(threshold: float = 19.5):
    """Returns O2 and rain sensor enclosure alerts.

    Parameters
    ----------
    threshold
        The threshold for the O2 level, as a percentage. If any of the sensors is
        below this value, the alert is raised.

    Returns
    -------
    enclosure_alerts_dict
        A dictionary of alerts. E.g.,
        ``{'o2_spec_room': False, 'o2_util_room': False, 'rain_sensor_alarm': True}``.

    """

    async with CluClient() as client:
        status = await client.send_command(
            "lvmecp",
            "status",
            internal=True,
            time_limit=5,
        )

    if status.status.did_fail:
        raise ValueError("Failed retrieving status from ECP.")

    registers = status.replies.get("registers")

    safety_labels = status.replies.get("safety_status_labels")
    door_alert = not ("DOOR_CLOSED" in safety_labels and "DOOR_LOCKED" in safety_labels)

    enclosure_alerts_dict = {
        "o2_spec_room": status.replies.get("o2_percent_spectrograph") < threshold,
        "o2_util_room": status.replies.get("o2_percent_utilities") < threshold,
        "rain_sensor_alarm": registers["rain_sensor_alarm"],
        "door_alert": door_alert,
    }

    return enclosure_alerts_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-12-03
# @Filename: general.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import functools
from datetime import datetime, timedelta

from typing import Any

import psycopg
import psycopg.sql

from lvmapi import config


__all__ = ["timed_cache", "get_db_connection", "insert_to_database"]


def timed_cache(seconds: float):
    """A cache decorator that expires after a certain time.

    Modified from https://gist.github.com/Morreski/c1d08a3afa4040815eafd3891e16b945

    This only works for synchronous functions. For asynchronous functions, use
    ``aiocache.cached``.

    Parameters
    ----------
    seconds
        Number of seconds after which the cache is cleared.

    """

    def _wrapper(f):
        update_delta = timedelta(seconds=seconds)
        next_update = datetime.utcnow() + update_delta

        # Apply @lru_cache to f with no cache size limit
        f = functools.lru_cache(None)(f)

        @functools.wraps(f)
        def _wrapped(*args, **kwargs):
            nonlocal next_update
            now = datetime.utcnow()
            if now >= next_update:
                f.cache_clear()
                next_update = now + update_delta
            return f(*args, **kwargs)

        return _wrapped

    return _wrapper


def get_db_connection():
    """Returns a connection to the database."""

    uri = config["database.uri"]

    return psycopg.AsyncConnection.connect(uri)


async def insert_to_database(
    table_name: str,
    data: list[dict[str, Any]],
    columns: list[str] | None = None,
):
    """Inserts data into the database.

    Parameters
    ----------
    table_name
        The table in the database where to insert the data. Can be in the format
        ``schema.table_name``.
    data
        The data to ingest, as a list of dictionaries in which each dictionary
        is a mapping of column name in ``table`` to the value to ingest.
    columns
        A list of table columns. If not passed, the column names are inferred from
        the first element in the data. In this case you must ensure that all the
        elements in the data contain entries for all the columns (use :obj:`None`
        to fill missing data).

    """

    if len(data) == 0:
        return

    columns = columns or list(data[0].keys())

    schema: str | None
    table: psycopg.sql.Identifier
    if "." in table_name:
        schema, table_name = table_name.split(".")
        table = psycopg.sql.Identifier(schema, table_name)
    else:
        table = psycopg.sql.Identifier(table_name)

    columns_sql = [psycopg.sql.Identifier(col) for col in columns]

    column_placeholders = ("{}, " * len(columns))[0:-2]
    values_placeholders = ("%s, " * len(columns))[0:-2]

    query = psycopg.sql.SQL(
        "INSERT INTO {} ("
        + column_placeholders
        + ") VALUES ("
        + values_placeholders
        + ");"
    ).format(
        table,
        *columns_sql,
    )

    async with await get_db_connection() as aconn:
        async with aconn.cursor() as acursor:
            for row in data:
                values = [row.get(col, None) for col in columns]
                await acursor.execute(query, values)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-07-25
# @Filename: tasks.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import Any

from fastapi import APIRouter
from pydantic import BaseModel
from taskiq_redis.exceptions import ResultIsMissingError

from lvmapi.broker import broker


class TaskResult(BaseModel):
    """A model to represent the result of a task."""

    is_err: bool
    is_ready: bool
    log: str | None = None
    return_value: Any | None = None
    execution_time: float | None = None
    labels: dict | None = None
    error: str | None = None


router = APIRouter(prefix="/tasks", tags=["tasks"])


@router.get("/{task_id}/ready")
async def task_ready(task_id: str) -> bool:
    """Returns whether a tasks has finished running."""

    return await broker.result_backend.is_result_ready(task_id)


@router.get("/{task_id}/result")
async def task_result(task_id: str):
    """Returns whether a tasks has finished running."""

    try:
        result = await broker.result_backend.get_result(task_id)
    except ResultIsMissingError:
        return TaskResult(
            is_ready=False,
            is_err=True,
            error="Task not found or result not ready.",
        )

    error = result.error if result.error is None else str(result.error)

    return TaskResult(is_ready=True, **result.dict(exclude={"error"}), error=error)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-06-14
# @Filename: macros.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import Annotated

from fastapi import APIRouter, HTTPException, Query, Request

from lvmapi.auth import AuthDependency
from lvmapi.tasks import cleanup_task, shutdown_task


router = APIRouter(prefix="/macros", tags=["macros"])


@router.get(
    "/shutdown",
    dependencies=[AuthDependency],
    summary="Runs the shutdown macro",
)
async def route_get_shutdown(
    disable_overwatcher: Annotated[
        bool,
        Query(description="Disables the Overwatcher after closing the dome."),
    ],
) -> str:
    """Schedules an emergency shutdown of the enclosure and telescopes."""

    task = await shutdown_task.kiq(disable_overwatcher=disable_overwatcher)
    return task.task_id


@router.get(
    "/cleanup",
    dependencies=[AuthDependency],
    summary="Runs the cleanup macro",
)
async def route_get_cleanup(
    readout: Annotated[
        bool,
        Query(
            description="If the spectrographs are idle and with a readout pending, "
            "reads the spectrographs."
        ),
    ] = True,
) -> str:
    """Runs the cleanup recipe as a task."""

    task = await cleanup_task.kiq(readout=readout)
    return task.task_id


@router.get("/shutdown_from_dupont", include_in_schema=False)
async def route_get_shutdown_from_dupont(request: Request) -> str:
    """Schedules an emergency shutdown of the enclosure and telescopes.

    This is a special version of the shutdown macro that does not require
    authentication but that will fail unless the request comes from the
    DuPont telescope.

    """

    host_valid = request.client and request.client.host == "10.8.38.21"
    dupont_valid = request.headers.get("x-real-ip", "") == "10.8.68.201"

    if not host_valid or not dupont_valid:
        raise HTTPException(status_code=401, detail="Invalid request. Not authorised.")

    task = await shutdown_task.kiq()
    return task.task_id



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-03-26
# @Filename: weather.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import Annotated

import polars
from fastapi import APIRouter, HTTPException, Query

from lvmapi.tools.weather import get_weather_data


router = APIRouter(prefix="/weather", tags=["weather"])


@router.get("/", summary="Weather report")
@router.get("/report", summary="Weather report")
async def route_get_weather(
    station: Annotated[
        str,
        Query(description="The weather station to query"),
    ] = "DuPont",
    start_time: Annotated[
        str | None,
        Query(description="Start time for the query. ISO format."),
    ] = None,
    end_time: Annotated[
        str | None,
        Query(description="End time for the query. ISO format."),
    ] = None,
    delta_time: Annotated[
        float,
        Query(
            description="Look-back number of seconds. "
            "Ignored if start_time/end_time are defined."
        ),
    ] = 3600,
    last: Annotated[
        bool,
        Query(description="Return only the last entry."),
    ] = False,
) -> list[dict]:
    """Returns the weather report from a weather station."""

    if start_time and end_time:
        df = await get_weather_data(
            station=station,
            start_time=start_time,
            end_time=end_time,
        )
    elif any([start_time, end_time]) and not all([start_time, end_time]):
        raise HTTPException(400, "start_time and end_time must be both defined.")
    else:
        df = await get_weather_data(station=station, start_time=delta_time)

    if last and len(df) > 0:
        df = df.tail(1)

    return df.with_columns(ts=polars.col.ts.dt.to_string("%FT%X")).to_dicts()
